# -*- coding: UTF-8 -*-
"""
dtaidistance.dtw_visualisation
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW) visualisations.

:author: Wannes Meert
:copyright: Copyright 2017-2024 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import os
import logging

from . import util_numpy
from . import innerdistance


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
except ImportError:
    np = None


logger = logging.getLogger("be.kuleuven.dtai.distance")

from . import dtw
try:
    from . import dtw_c
except ImportError:
    # logger.info('C library not available')
    dtw_c = None

try:
    from tqdm import tqdm
except ImportError:
    logger.info('tqdm library not available')
    tqdm = None


def test_without_visualization():
    if "DTAIDISTANCE_TESTWITHOUTVIZ" in os.environ and os.environ["DTAIDISTANCE_TESTWITHOUTVIZ"] == "1":
        return True
    return False


def plot_warp(from_s, to_s, new_s, path, filename=None, fig=None, axs=None):
    """Plot the warped sequence and its relation to the original sequence
    and the target sequence.

    :param from_s: From sequence.
    :param to_s: To sequence.
    :param new_s: Warped version of from sequence.
    :param path: Optimal warping path.
    :param filename: Filename path (optional).
    :param fig: Matplotlib Figure object
    :param axs: Array of Matplotlib axes.Axes objects (length == 3)
    :return: Figure, list[Axes]
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    if fig is None and axs is None:
        fig, axs = plt.subplots(nrows=3, ncols=1, sharex='all', sharey='all')
    elif fig is None or axs is None:
        raise TypeError(f'The fig and axs arguments need to be both None or both instantiated.')
    axs[0].plot(from_s, label="From")
    axs[0].legend()
    axs[1].plot(to_s, label="To")
    axs[1].legend()
    lines = []
    line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    for r_c, c_c in path:
        if r_c < 0 or c_c < 0:
            continue
        con = ConnectionPatch(xyA=[r_c, from_s[r_c]], coordsA=axs[0].transData,
                              xyB=[c_c, to_s[c_c]], coordsB=axs[1].transData, **line_options)

        lines.append(con)
    axs[2].plot(new_s, label="From-warped")
    axs[2].legend()
    for i in range(len(to_s)):
        con = ConnectionPatch(xyA=[i, to_s[i]], coordsA=axs[1].transData,
                              xyB=[i, new_s[i]], coordsB=axs[2].transData, **line_options)
        lines.append(con)
    for line in lines:
        fig.add_artist(line)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, axs = None, None
    return fig, axs


def plot_warping(s1, s2, path, filename=None, fig=None, axs=None,
                 series_line_options=None, warping_line_options=None,
                 start_on_curve=True, color_misalignment=False, tick_kwargs=None):
    """Plot the optimal warping between two sequences.

    :param s1: From sequence.
    :param s2: To sequence.
    :param path: Optimal warping path.
        Can be computed with methods like :meth:`dtaidistance.dtw.warping_path`.
    :param filename: Filename path (optional).
    :param fig: Matplotlib Figure object
    :param axs: Array of Matplotlib axes.Axes objects (length == 2)
    :param series_line_options: Dictionary of options to pass to matplotlib plot
        None will not pass any options
    :param warping_line_options: Dictionary of options to pass to matplotlib ConnectionPatch
        None will use {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    :param start_on_curve: Start the line on the curve. If false, start on the Y-axis.
    :param color_misalignment: Color delayed, on-time, or early based on the misalignment
        function.
    :return: Figure, list[Axes]
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    if fig is None and axs is None:
        if color_misalignment:
            fig, axs = plt.subplots(nrows=2, ncols=1, sharex='all', sharey='all', gridspec_kw = {'hspace': 1.0})
        else:
            fig, axs = plt.subplots(nrows=2, ncols=1, sharex='all', sharey='all')
    elif fig is None or axs is None:
        raise TypeError(f'The fig and axs arguments need to be both None or both instantiated.')
    if series_line_options is None:
        series_line_options = {}
    axs[0].plot(s1, **series_line_options)
    axs[1].plot(s2, **series_line_options)
    plt.tight_layout()
    s1_min, _ = axs[0].get_ylim()
    _, s2_max = axs[1].get_ylim()
    lines = []
    if warping_line_options is None:
        warping_line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    if color_misalignment:
        warping_line_options['color'] = 'yellow'
        g_np = 0
    else:
        g_np = None
    for r_c, c_c in path:
        if r_c < 0 or c_c < 0:
            continue
        if start_on_curve:
            s1_y = s1[r_c]
            s2_y = s2[c_c]
        else:
            s1_y = s1_min
            s2_y = s2_max
        if color_misalignment:
            # Based on:
            # D. K. Urribarri, M. L. Larrea, S. M. Castro, and E. Puppo.
            # Overview+ detail visual comparison of karate motion captures.
            # In Computer Science–CACIC 2019: 25th Argentine Congress of Computer
            # Science, CACIC 2019, Springer, 2020.
            g_n = c_c - r_c  # misalignment function
            if g_np == g_n:  # on-time
                warping_line_options['color'] = '#E4D576'  # yellow
            elif g_np < g_n:  # delayed
                warping_line_options['color'] = '#F27D7D'  # red
            else:  # ahead
                warping_line_options['color'] = '#7DBDF2'  # blue
            g_np = g_n
        con = ConnectionPatch(xyA=[r_c, s1_y], coordsA=axs[0].transData,
                              xyB=[c_c, s2_y], coordsB=axs[1].transData, **warping_line_options)
        lines.append(con)
    for line in lines:
        fig.add_artist(line)
    if tick_kwargs is not None:
        for ax in axs:
            ax.tick_params(**tick_kwargs)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, axs = None, None
    return fig, axs


def plot_warping_single_ax(s1, s2, path, filename=None, fig=None, ax=None):
    """Plot the optimal warping between to sequences.

    :param s1: From sequence.
    :param s2: To sequence.
    :param path: Optimal warping path.
    :param filename: Filename path (optional).
    :param fig: Matplotlib Figure object
    :param ax: Matplotlib axes.Axes object
    :return: Figure, Axes
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    if fig is None and ax is None:
        fig, ax = plt.subplots()
    elif fig is None or ax is None:
        raise TypeError(f'The fig and ax arguments need to be both None or both instantiated.')
    ax.plot(s1)
    ax.plot(s2)
    lines = []
    line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    for r_c, c_c in path:
        if r_c < 0 or c_c < 0:
            continue
        con = ConnectionPatch(xyA=[r_c, s1[r_c]], coordsA=ax.transData,
                              xyB=[c_c, s2[c_c]], coordsB=ax.transData, **line_options)
        lines.append(con)
    for line in lines:
        fig.add_artist(line)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax


def path_slice(path, rb=None, re=None, cb=None, ce=None):
    path2 = []
    for t in path:
        if rb is not None and t[0] < rb:
            continue
        if cb is not None and t[1] < cb:
            continue
        if re is not None and t[0] > (re - 1):
            continue
        if ce is not None and t[1] > (ce - 1):
            continue
        path2.append((t[0] - rb, t[1] - cb))
    return path2


def plot_warpingpaths(s1, s2, paths, cost_matrix = None, path=None, filename=None,
                      shownumbers=False, showlegend=False, showtotaldist=True,
                      figure=None, path_kwargs = None, matshow_kwargs=None,
                      plot_kwargs = None,
                      includes_zero=True, tick_kwargs=None, show_diagonal=False):
    """Plot the warping paths matrix.

    :param s1: Series 1
    :param s2: Series 2
    :param paths: Warping paths matrix
    :param cost_matrix: Cost matrix, if it is not None, it will be plotted instead of the accummulated cost matrix ('paths').
    :param path: Path to draw (typically this is the best path)
    :param filename: Filename for the image (optional)
    :param shownumbers: Show distances also as numbers
    :param showlegend: Show colormap legend
    :param figure: Matplotlib Figure object
    :param path_kwargs: kwargs for the path plot
    :param matshow_kwargs: kwargs for the matshow plot

    :return: Figure, Axes
    """
    try:
        from matplotlib import pyplot as plt
        from matplotlib import gridspec
        from matplotlib.ticker import FuncFormatter
    except ImportError:
        logger.error("The plot_warpingpaths function requires the matplotlib package to be installed.")
        return
    ratio = max(len(s1), len(s2))
    min_y = min(np.min(s1), np.min(s2))
    max_y = max(np.max(s1), np.max(s2))
    if plot_kwargs is None:
        plot_kwargs = {'linestyle': '-'}

    if figure is None:
        fig = plt.figure(figsize=(10, 10), frameon=True)
    else:
        fig = figure
    size = fig.get_size_inches()
    if showlegend:
        grows = 3
        gcols = 3
        height_ratios = [1, 6, 1]
        width_ratios = [1, 6, 1]
    else:
        grows = 2
        gcols = 2
        if len(s1) == len(s2):
            height_ratios = [1, 6]
            width_ratios = [1, 6]
        elif len(s1) < len(s2):
            height_ts = size[1] / 7
            height_ratios = [height_ts, size[1] - height_ts]
            width_ratios = [height_ts, size[0] - height_ts]
        elif len(s1) > len(s2):
            width_ts = size[0] / 7
            height_ratios = [width_ts, size[1] - width_ts]
            width_ratios = [width_ts, size[0] - width_ts]
    gs = gridspec.GridSpec(grows, gcols, wspace=1, hspace=1,
                           left=0, right=1.0, bottom=0, top=1.0,
                           height_ratios=height_ratios,
                           width_ratios=width_ratios)
    max_s2_x = np.max(s2)
    max_s2_y = len(s2) - 1
    max_s1_y = len(s1) - 1
    y_ratio2 = (np.max(s1) - np.min(s1)) / (np.max(s2) - np.min(s2))
    y_ratio1 = min(1.0, 1.0 / y_ratio2)
    y_ratio2 = min(1.0, y_ratio2)

    if path is None and includes_zero is True:
        p = dtw.best_path(paths)
    elif type(path) is int and path == -1:
        p = None
    else:
        p = path

    def format_fn2_x(tick_val, tick_pos):
        return max_s2_x - tick_val

    def format_fn2_y(tick_val, tick_pos):
        return int(max_s2_y - tick_val)

    # Warping path
    ax3 = fig.add_subplot(gs[1, 1])
    # ax3.set_aspect(1)
    if matshow_kwargs is None:
        matshow_kwargs = {}
    if cost_matrix is not None:
        img = ax3.matshow(cost_matrix[1:, 1:], aspect='equal', **matshow_kwargs)
    else:
        if includes_zero:
            if isinstance(includes_zero, innerdistance.StepsType):
                print(f'{paths.shape=}')
                inf_rows, inf_cols = includes_zero.inf_rows_cols()
            else:
                inf_rows, inf_cols = 1, 1
            img = ax3.matshow(paths[inf_rows:, inf_cols:],
                              aspect='equal', **matshow_kwargs)
        else:
            img = ax3.matshow(paths, aspect='equal', **matshow_kwargs)
    if show_diagonal:
        ax3.plot([0, len(s2)], [0, len(s1)], '-', color='white', alpha=0.4)
    # ax3.grid(which='major', color='w', linestyle='-', linewidth=0)
    # ax3.set_axis_off()
    if p is not None:
        if type(p) is list:
           py, px = zip(*p)
        else:
            py = p[:, 0]
            px = p[:, 1]

        path_kwargs = {'color':"red"} if path_kwargs is None else path_kwargs
        ax3.plot(px, py, ".-", **path_kwargs)

    # ax3.xaxis.set_major_locator(plt.NullLocator())
    # ax3.yaxis.set_major_locator(plt.NullLocator())
    if shownumbers:
        for r in range(1, paths.shape[0]):
            for c in range(1, paths.shape[1]):
                ax3.text(c - 1, r - 1, "{:.2f}".format(paths[r, c]), ha='center', va='center')
    ax3.xaxis.set_ticks_position('bottom')
    ax3.yaxis.set_ticks_position('right')

    # Time series on top axis
    ax1 = fig.add_subplot(gs[0, 1])
    ax1.set_ylim([min_y, max_y])
    ax1.set_axis_off()
    ax1.xaxis.tick_top()
    if 'color' in plot_kwargs:
        ax1.plot(range(len(s2)), s2, **plot_kwargs)
    else:
        ax1.plot(range(len(s2)), s2, color='#ff7f0e', **plot_kwargs)
    ax1.set_xlim([-0.5, len(s2) - 0.5])
    ax1.xaxis.set_major_locator(plt.NullLocator())
    ax1.yaxis.set_major_locator(plt.NullLocator())

    # Time series on left axis
    ax2 = fig.add_subplot(gs[1, 0])
    ax2.set_xlim([-max_y, -min_y])
    ax2.set_axis_off()
    # ax2.xaxis.set_major_formatter(FuncFormatter(format_fn2_x))
    # ax2.yaxis.set_major_formatter(FuncFormatter(format_fn2_y))
    ax2.xaxis.set_major_locator(plt.NullLocator())
    ax2.yaxis.set_major_locator(plt.NullLocator())
    if 'color' in plot_kwargs:
        ax2.plot(-s1, range(len(s1)), **plot_kwargs)
    else:
        ax2.plot(-s1, range(len(s1)), color='#1f77b4', **plot_kwargs)
    ax2.set_ylim([-0.5, len(s1) - 0.5])
    ax2.invert_yaxis()

    # for ax in [ax0, ax1, ax2, ax3]:
    #     for spine in ax.spines.values():
    #         spine.set_visible(True)
    #         spine.set_edgecolor('green')  # Set border color
    #         spine.set_linewidth(2)  # Set border thickness

    gs.tight_layout(fig, pad=1.0, h_pad=1.0, w_pad=1.0)
    # fig.subplots_adjust(hspace=0, wspace=0)

    ax0 = fig.add_subplot(gs[0, 0])
    ax0.set_axis_off()
    if p is not None and showtotaldist:
        ax0.text(0, 0, "Dist = {:.4f}".format(paths[p[-1][0] + 1, p[-1][1] + 1]))
    ax0.xaxis.set_major_locator(plt.NullLocator())
    ax0.yaxis.set_major_locator(plt.NullLocator())

    if showlegend:
        # ax4 = fig.add_subplot(gs[0:, 2])
        ax4 = fig.add_axes([0.9, 0.25, 0.015, 0.5])
        fig.colorbar(img, cax=ax4)

    # Align the subplots:
    if len(s1) != len(s2):
        # bounds = (xmin, ymin, width, height)
        ax1pos = ax1.get_position().bounds  # top ts
        ax2pos = ax2.get_position().bounds  # left ts
        ax3pos = ax3.get_position().bounds  # warping path
        left = ax3pos[0]
        bottom = ax3pos[1]
        width = ax3pos[2]
        height = ax3pos[3]
        tsheight = min(ax1pos[3], ax2pos[2], height / 6)
        dist_between_y = ax1pos[1] - (ax2pos[1] + ax2pos[3])
        dist_between_x = ax1pos[0] - ax2pos[0]  # Tts.left - Lts.left
        # set_position([left, bottom, width, height])
        ax0.set_position((left - 1.1 * tsheight, bottom+height+dist_between_y, tsheight, tsheight))
        ax1.set_position((left, bottom+height+dist_between_y, width, tsheight))
        ax2.set_position((left - 1.1 * tsheight, bottom, tsheight, height))

    ax = fig.axes
    if tick_kwargs is not None:
        ax3.tick_params(**tick_kwargs)
    # ax3.spines['bottom'].set_linewidth(2.0)
    # ax3.spines['right'].set_linewidth(2.0)
    for spine in ax3.spines.values():
        spine.set_linewidth(2)  # Increase axis thickness
        # spine.set_color('black')  # Ensure it's visible
    if filename:
        if type(filename) != str:
            filename = str(filename)
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax


def plot_warpingpaths_addpath(ax, path):
    py, px = zip(*path)
    ax3 = ax[0]
    ax3.plot(px, py, ".-", color="red", markersize=2)


def plot_matrix(distances, shownumbers=False, filename=None, fig=None, ax=None):
    from matplotlib import pyplot as plt

    if ax is None and fig is None:
        if shownumbers:
            figsize = (15, 15)
        else:
            figsize = None
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
    elif fig is None or ax is None:
        raise TypeError(f'The fig and ax arguments need to be both None or both instantiated.')
    else:
        fig = None

    ax.xaxis.set_ticks_position('top')
    ax.yaxis.set_ticks_position('both')

    im = ax.imshow(distances)
    idxs_y = [str(i) for i in range(distances.shape[0])]
    idxs_x = [str(i) for i in range(distances.shape[1])]
    # Show all ticks
    ax.set_xticks(np.arange(len(idxs_x)))
    ax.set_xticklabels(idxs_x)
    ax.set_yticks(np.arange(len(idxs_y)))
    ax.set_yticklabels(idxs_y)

    ax.set_title("Distances between series", pad=30)

    if shownumbers:
        for i in range(len(idxs_y)):
            for j in range(len(idxs_x)):
                if not np.isinf(distances[i, j]):
                    l = "{:.2f}".format(distances[i, j])
                    ax.text(j, i, l, ha="center", va="center", color="w")

    if filename:
        if type(filename) != str:
            filename = str(filename)
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax


def plot_average(s1, s2, avg, path1, path2, filename=None, fig=None, ax=None):
    """Plot how s1 and s2 relate to the avg.

    :param s1: Seq 1.
    :param s2: Seq 2.
    :param path: Average sequence.
    :param filename: Filename path (optional).
    :param fig: Matplotlib Figure object
    :param ax: Matplotlib axes.Axes object
    :return: Figure, axes.Axes
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    if fig is None and ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1, sharex='all', sharey='all')
    elif fig is None or ax is None:
        raise TypeError(f'The fig and axs arguments need to be both None or both instantiated.')
    ax.plot(s1, color='blue', alpha=0.5)
    ax.plot(s2, color='blue', alpha=0.5)
    ax.plot(avg, color='orange', linestyle='dashed', alpha=0.5)
    # plt.tight_layout()
    # lines = []
    # line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    # for r_c, c_c in path:
    #     if r_c < 0 or c_c < 0:
    #         continue
    #     con = ConnectionPatch(xyA=[r_c, s1[r_c]], coordsA=ax[0].transData,
    #                           xyB=[c_c, s2[c_c]], coordsB=ax[1].transData, **line_options)
    #     lines.append(con)
    # for line in lines:
    #     fig.add_artist(line)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax



# -*- coding: UTF-8 -*-
"""
dtaidistance.innerdistance
~~~~~~~~~~~~~~~~~~~~~~~~~~

Inner (or local) distances for DTW and ED

We make the distinction between three operations:

- The `inner_dist`: The inner distance (or local distance, or cost) is
    the function that expresses the similarity between two values in the
    series in the cost space.
- The `result`: The transformation to change from the cost space to
    the (global) distance space.
- The `inner_val`: The transformation to change from the (global) distance
    space to the cost space.

:author: Wannes Meert
:copyright: Copyright 2023-2025 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import math
import logging
from enum import Enum

from . import util
from . import util_numpy

# The available inner distances, with their integer identifier are:
# - 0: squared Euclidean (for univariate series this is squared difference)
# - 1: Euclidean (for univariate series this his absolute difference)

try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    DTYPE = np.double
    argmin = np.argmin
    argmax = np.argmax
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    argmin = util.argmin
    argmax = util.argmax
    array_min = min
    array_max = max


logger = logging.getLogger("be.kuleuven.dtai.distance")
default = 'squared euclidean'


class InnerDistType(util.DDType):
    SQEUCLIDEAN = "squared euclidean"  # 0
    EUCLIDEAN = "euclidean"  # 1


class SquaredEuclidean:
    """Squared Euclidean inner distance for univariate series.
    This is identical to squared difference.
    """

    @staticmethod
    def inner_dist(x, y):
        return (x - y) ** 2

    @staticmethod
    def result(x):
        if np is not None and isinstance(x, np.ndarray):
            return np.sqrt(x)
        return math.sqrt(x)

    @staticmethod
    def inner_val(x):
        return x*x


class SquaredEuclideanNdim:
    """Squared Euclidean inner distance for multivariate series."""

    @staticmethod
    def inner_dist(x, y):
        return np.sum((x - y) ** 2)

    @staticmethod
    def result(x):
        return np.sqrt(x)

    @staticmethod
    def inner_val(x):
        return x * x


class Euclidean:
    """Euclidean inner distance for univariate series.
    This is identical to absoluate difference.
    """

    @staticmethod
    def inner_dist(x, y):
        return abs(x - y)

    @staticmethod
    def result(x):
        return x

    @staticmethod
    def inner_val(x):
        return x


class EuclideanNdim:
    """Squared Euclidean inner distance for multivariate series."""

    @staticmethod
    def inner_dist(x, y):
        return np.sqrt(np.sum(np.power(x - y, 2)))

    @staticmethod
    def result(x):
        return x

    @staticmethod
    def inner_val(x):
        return x


class CustomInnerDist:
    """API to create your own custom inner (local)distance."""

    @staticmethod
    def inner_dist(x, y):
        """The cost (or inner distance, or local distance) between two points in the series.

        For n-dimensional data, the two arguments x and y will be vectors.
        Otherwise, they are scalars.

        For example, for default DTW this would be the Squared Euclidean
        distance: (a-b)**2.
        """
        raise Exception("Function not defined")

    @staticmethod
    def result(x):
        """The transformation applied to the sum of all inner distances.
        We also refer to this method as cost-to-dist.
        The inverse operation of `inner_val`.

        The variable x can be both a single number as a matrix.

        For example, for default DTW, which uses Squared Euclidean, this
        would be: sqrt(d). Because d = (a_0-b_0)**2 + (a_1-b_1)**2 ...
        """
        raise Exception("Function not defined")

    @staticmethod
    def inner_val(x):
        """The transformation applied to input settings like max_step.
        We also refer to this method as dist-to-cost.
        The inverse operation of `result`."""
        raise Exception("Function not defined")


def inner_dist_cls(inner_dist="squared euclidean", use_ndim=False):
    if np is not None:
        if np.issubdtype(type(inner_dist), np.integer):
            inner_dist = int(inner_dist)
    if type(inner_dist) in [str, int]:
        inner_dist = InnerDistType.wrap(inner_dist)

    if inner_dist == InnerDistType.SQEUCLIDEAN:
        if use_ndim:
            use_cls = SquaredEuclideanNdim
        else:
            use_cls = SquaredEuclidean
    elif inner_dist == InnerDistType.EUCLIDEAN:
        if use_ndim:
            use_cls = EuclideanNdim
        else:
            use_cls = Euclidean
    elif hasattr(inner_dist, 'inner_dist') and hasattr(inner_dist, 'result'):
        use_cls = inner_dist
    else:
        raise AttributeError(f"Unknown value for argument inner_dist: {inner_dist}")
    return use_cls


def inner_dist_fns(inner_dist="squared euclidean", use_ndim=False):
    """Return the inner distance (or local distance) functions and transformations.

    :param inner_dist: Type of inner_dist
    :param use_ndim: Use multivariate or not
    :return: (inner_dist, cost2dist, dist2cost)
    """
    use_cls = inner_dist_cls(inner_dist, use_ndim)
    return use_cls.inner_dist, use_cls.result, use_cls.inner_val


def to_c(inner_dist):
    if hasattr(inner_dist, 'inner_dist') and hasattr(inner_dist, 'result'):
        raise AttributeError('Custom inner distance functions are not supported for the fast C implementation')
    inner_dist = InnerDistType.wrap(inner_dist)
    return inner_dist.to_int()


class StepsType(util.DDType):
    # Based on L. Rabiner and B.-H. Juang. Fundamentals of speech recognition.
    # Prentice-Hall, Inc., 1993.
    TYPEI = "TypeI"  # 0
    TYPEIII = "TypeIII"  # 1
    DIAGONAL = "Diagonal"  # 2

    def steps(self):
        return stepstype_steps[self]

    def inf_rows_cols(self):
        steps_rows, steps_cols = zip(*self.steps())
        inf_rows = max(steps_rows)
        inf_cols = max(steps_cols)
        return inf_rows, inf_cols


stepstype_steps = {
    StepsType.TYPEI: ((1, 1), (0, 1), (1, 0)),  # diagonal, go left, go up
    StepsType.TYPEIII: ((1, 1), (1, 2), (2, 1)),
    StepsType.DIAGONAL: ((1, 1),)
}



import numpy as np
import scipy as sp


def pattern1(x, x0=4, c=0.5, a=0, x1=22, d=1, r=0.0, rs=3980, w0=1, x2=25, w2=0):
    """A time series generator that simulates simple patterns in transient
    systems. It is a rising function with an overshoot and a short, one
    cycle of sine-like behavior.

    :param x0: Location of overshoot peak
    :param w0: Width of overshoot peak
    :param c: Height of converged level
    :param a: Add to height of overshoot peak
    :param x1: Location of sine wave
    :param d: Spread of sine wave
    :param r: Random noise level
    :param rs: Random seed
    """

    # Overshoot
    y = (sp.special.dawsn((x - x0)*w0) + c) * np.heaviside(x - x0, 0)
    y += (np.exp(x*w0) / (np.exp(x0*w0) / c)) * np.heaviside(x0 - x, 0)
    if a > 0:
        y += sp.stats.norm.pdf(x - x0) * a
    # Sine wave
    xd = 3
    y0 = sp.special.dawsn(xd)
    d = sp.special.dawsn((x - x1) * d)
    idx1 = d > y0
    idx2 = d < -y0
    d[idx1] = d[idx1] - y0
    d[idx2] = d[idx2] + y0
    d[~idx1 & ~idx2] = 0
    y += np.heaviside(x - x1 + xd, 0) * np.heaviside(x1 - x + xd, 0) * d / 2
    # Dip
    if w2 > 0:
        rv = sp.stats.norm(loc=x2, scale=w2/10)
        peak = rv.pdf(x2)
        y -= rv.pdf(x)/peak*c
    # Random noise
    if r > 0:
        np.random.seed(rs)
        y += np.random.random(y.shape) * r
    return y


def pattern2(length_of_ts, starting_index_of_wave, length_of_wave):
    return np.concatenate((np.zeros(starting_index_of_wave),
                           np.sin(np.linspace(0, 2 * np.pi, length_of_wave)),
                           np.zeros(length_of_ts - starting_index_of_wave - length_of_wave)))


def ts_under_pattern1(params_of_cluster=None, params_of_new_times=None):
    # from dtaidistance.benchmarks.synthetic import pattern1
    ys = []
    ys_new = []
    x = np.linspace(0, 30, num=200)
    x00 = 7
    x1 = 22
    if params_of_cluster == None:
        params_of_cluster = [
            (0, 0.5, 1., 1),
            (-0.7, 0.5, 1., 1),
            (-0.3, 0.5, 2., 2),
            (0.7, 0.5, 1., 2),
            (0.4, 0.5, 1., 3)
        ]

    if params_of_new_times == None:
        params_of_new_times = [
            (5, 0.5, 1, 1)
        ]

    for x0d, c, a, d in params_of_cluster:
        y = pattern1(x, x00 + x0d, c, a, x1, d)
        ys.append(y)

    for x0d, c, a, d in params_of_new_times:
        y = pattern1(x, x00 + x0d, c, a, x1, d)
        ys_new.append(y)

    return x, ys, ys_new




import logging
import sys, os
import random
from pathlib import Path
sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))

import pytest

from dtaidistance.preprocessing import differencing, derivative
from dtaidistance import util_numpy

logger = logging.getLogger("be.kuleuven.dtai.distance")
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")
scipyonly = pytest.mark.skipif("util_numpy.test_without_scipy()")


@numpyonly
@scipyonly
def test_differencing():
    with util_numpy.test_uses_numpy() as np:
        series = np.array([0.1, 0.3, 0.2, 0.1] * 3)
        series = differencing(series, smooth=0.1)
        np.testing.assert_array_almost_equal(
            series, np.array([0.02217,  0.010307,  0.002632,  0.001504,  0.001629, -0.000457,
                              -0.001698, -0.001238, -0.004681, -0.014869, -0.026607]))

        series = np.array([[0.1, 0.3, 0.2, 0.1] * 3])
        series = differencing(series, smooth=0.1)
        np.testing.assert_array_almost_equal(
            series, np.array([[0.02217,  0.010307,  0.002632,  0.001504,  0.001629, -0.000457,
                              -0.001698, -0.001238, -0.004681, -0.014869, -0.026607]]))


@numpyonly
@scipyonly
def test_derivative():
    with util_numpy.test_uses_numpy() as np:
        def der2(qim, qip):
            return (qip-qim)

        def der3(qim, qi, qip):
            return ((qi-qim) + ((qip-qim)/2))/2

        series = np.array([0.1, 0.15, 0.27, 0.3, 0.26, 0.24, 0.2, 0.1])
        seriesds = [der2(*series[0:2])] + \
                   [der3(qim, qi, qip) for qim, qi, qip in zip(series[:-2], series[1:-1], series[2:])] + \
                   [der2(*series[-2:])]
        seriesd = derivative(series, smooth=None)
        # import matplotlib.pyplot as plt
        # fig, axs = plt.subplots(3, 1)
        # axs[0].plot(series)
        # for i, d in enumerate(seriesd):
        #     x = [i-1, i+1]
        #     y = [series[i]-d, series[i]+d]
        #     axs[0].plot(x, y, color="green", alpha=0.5)
        #     d = seriesds[i]
        #     y = [series[i] - d, series[i] + d]
        #     axs[0].plot(x, y, color="orange", alpha=0.5)
        # axs[1].plot(seriesd)
        # axs[2].plot(differencing(series))
        # plt.show()
        np.testing.assert_array_almost_equal(seriesd[1:-1], np.array([0.0675, 0.0975, 0.0125, -0.035, -0.025, -0.055]))
        np.testing.assert_array_almost_equal(seriesd, seriesds)


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    test_differencing()




from dtaidistancec_globals cimport seq_t, StepType, DDPath
from dtaidistancec_dtw cimport DTWSettings

cdef extern from "dd_loco.h":
    ctypedef struct LoCoSettings:
        Py_ssize_t window
        seq_t penalty
        Py_ssize_t psi_1b
        Py_ssize_t psi_2b
        bint only_triu
        seq_t gamma
        seq_t tau
        seq_t delta
        seq_t delta_factor
        StepType step_type

    LoCoSettings loco_settings_default()

    seq_t loco_warping_paths(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, LoCoSettings *settings)
    seq_t loco_warping_paths_typeI(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2,  LoCoSettings *settings)
    seq_t loco_warping_paths_ndim_typeI(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, int ndim,
                                        LoCoSettings *settings)
    seq_t loco_warping_paths_typeIII(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, LoCoSettings *settings)
    seq_t loco_warping_paths_ndim_typeIII(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, int ndim,
                                          LoCoSettings *settings)
    void loco_path_negativize(Py_ssize_t *path, Py_ssize_t length, seq_t *wps, Py_ssize_t l1, Py_ssize_t l2, int buffer, int inf_rows,
                              int inf_cols)

    void best_path_init(DDPath *a, size_t initialSize)
    void best_path_insert(DDPath *a, Py_ssize_t element)
    void best_path_free(DDPath *a)
    DDPath loco_best_path(seq_t *wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t r, Py_ssize_t c, int min_size, LoCoSettings *settings)

    void loco_wps_argmax(seq_t *wps, Py_ssize_t l, Py_ssize_t *idxs, int n)




from dtaidistancec_globals cimport seq_t, DDPath


cdef extern from "dd_dtw.h":
    ctypedef struct DTWSettings:
        Py_ssize_t window
        seq_t max_dist
        seq_t max_step
        Py_ssize_t max_length_diff
        seq_t penalty
        Py_ssize_t psi_1b
        Py_ssize_t psi_1e
        Py_ssize_t psi_2b
        Py_ssize_t psi_2e
        bint use_pruning
        int inner_dist

    ctypedef struct DTWBlock:
        Py_ssize_t rb
        Py_ssize_t re
        Py_ssize_t cb
        Py_ssize_t ce
        bint triu

    ctypedef struct DTWWps:
        Py_ssize_t ldiff
        Py_ssize_t ldiffr
        Py_ssize_t ldiffc
        Py_ssize_t window
        Py_ssize_t width
        Py_ssize_t length
        Py_ssize_t ri1
        Py_ssize_t ri2
        Py_ssize_t ri3
        Py_ssize_t overlap_left_ri
        Py_ssize_t overlap_right_ri
        seq_t max_step
        seq_t max_dist
        seq_t penalty

    DTWSettings dtw_settings_default()
    Py_ssize_t dtw_settings_wps_length(Py_ssize_t l1, Py_ssize_t l2, DTWSettings *settings)
    Py_ssize_t dtw_settings_wps_width(Py_ssize_t l1, Py_ssize_t l2, DTWSettings *settings)
    void dtw_settings_set_psi(Py_ssize_t psi, DTWSettings *settings)
    void dtw_print_settings(DTWSettings *settings)

    seq_t dtw_distance(seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2,
                          DTWSettings *settings)
    seq_t dtw_distance_ndim(seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, int ndim,
                               DTWSettings *settings)

    seq_t dtw_warping_paths(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, int l2,
                               bint return_dtw, bint keep_int_repr, bint psi_neg, DTWSettings *settings)
    seq_t dtw_warping_paths_ndim(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, int l2,
                                 bint return_dtw, bint keep_int_repr, bint psi_neg, int ndim, DTWSettings *settings)
    seq_t dtw_warping_paths_affinity(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, bint return_dtw,
                                     bint keep_int_repr, bint psi_neg, bint only_triu, seq_t gamma, seq_t tau, seq_t delta,
                                     seq_t delta_factor, DTWSettings *settings)
    seq_t dtw_warping_paths_affinity_ndim(seq_t *wps, seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, bint return_dtw,
                                          bint keep_int_repr, bint psi_neg, bint only_triu, int ndim, seq_t gamma, seq_t tau, seq_t delta,
                                          seq_t delta_factor, DTWSettings *settings)
    void dtw_expand_wps(seq_t *wps, seq_t *full, Py_ssize_t l1, Py_ssize_t l2, DTWSettings *settings)
    void dtw_expand_wps_slice(seq_t *wps, seq_t *full, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb, Py_ssize_t ce,
                              DTWSettings *settings)
    void dtw_expand_wps_affinity(seq_t *wps, seq_t *full, Py_ssize_t l1, Py_ssize_t l2, DTWSettings *settings)
    void dtw_expand_wps_slice_affinity(seq_t *wps, seq_t *full, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb,
                                       Py_ssize_t ce, DTWSettings *settings)
    bint dtw_wps_negativize_value(DTWWps *p, seq_t *wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t r, Py_ssize_t c)
    bint dtw_wps_positivize_value(DTWWps *p, seq_t *wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t r, Py_ssize_t c)
    void dtw_wps_positivize(DTWWps *p, seq_t *wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb, Py_ssize_t ce, bint intersection)
    void dtw_wps_negativize(DTWWps *p, seq_t *wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb, Py_ssize_t ce, bint intersection)
    Py_ssize_t dtw_wps_loc(DTWWps *p, Py_ssize_t r, Py_ssize_t c, Py_ssize_t l1, Py_ssize_t l2)
    Py_ssize_t dtw_wps_max(DTWWps * p, seq_t *wps, Py_ssize_t *r, Py_ssize_t *c, Py_ssize_t l1, Py_ssize_t l2)
    Py_ssize_t dtw_best_path(seq_t *wps, Py_ssize_t *i1, Py_ssize_t *i2, Py_ssize_t l1, Py_ssize_t l2,
                             DTWSettings *settings)
    Py_ssize_t dtw_best_path_affinity(seq_t *wps, Py_ssize_t *i1, Py_ssize_t *i2, Py_ssize_t l1, Py_ssize_t l2,
                                      Py_ssize_t s1, Py_ssize_t s2,
                                      DTWSettings *settings)
    Py_ssize_t dtw_best_path_prob(seq_t *wps, Py_ssize_t *i1, Py_ssize_t *i2, Py_ssize_t l1, Py_ssize_t l2,
                                  seq_t avg, DTWSettings *settings);
    seq_t dtw_warping_path(seq_t *from_s, Py_ssize_t from_l, seq_t* to_s, Py_ssize_t to_l,
                           Py_ssize_t *from_i, Py_ssize_t *to_i, Py_ssize_t *length_i, DTWSettings * settings)
    seq_t dtw_warping_path_ndim(seq_t *from_s, Py_ssize_t from_l, seq_t * to_s, Py_ssize_t to_l,
                                Py_ssize_t *from_i, Py_ssize_t *to_i, Py_ssize_t *length_i, int ndim, DTWSettings * settings)
    void dtw_srand(unsigned int seed)
    seq_t dtw_warping_path_prob_ndim(seq_t *from_s, Py_ssize_t from_l, seq_t* to_s, Py_ssize_t to_l,
                                     Py_ssize_t *from_i, Py_ssize_t *to_i, Py_ssize_t *length_i, seq_t avg, int ndim, DTWSettings * settings)
    DTWWps dtw_wps_parts(Py_ssize_t l1, Py_ssize_t l2, DTWSettings * settings)

    DDPath dtw_wph_sqeuc_typei(seq_t *f_s, Py_ssize_t f_l,
                               seq_t* t_s, Py_ssize_t t_l,
                               Py_ssize_t switch_to_full,
                               int ndim, DTWSettings * settings)

    seq_t ub_euclidean(seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2)
    seq_t ub_euclidean_ndim(seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, int ndim)
    seq_t lb_keogh(seq_t *s1, Py_ssize_t l1, seq_t *s2, Py_ssize_t l2, DTWSettings *settings)

    void dtw_dba_ptrs(seq_t **ptrs, Py_ssize_t nb_ptrs, Py_ssize_t* lengths,
                  seq_t *c, Py_ssize_t t, unsigned char *mask, int prob_samples, int ndim, DTWSettings *settings)
    void dtw_dba_matrix(seq_t *matrix, Py_ssize_t nb_rows, Py_ssize_t nb_cols,
                    seq_t *c, Py_ssize_t t, unsigned char *mask, int prob_samples, int ndim, DTWSettings *settings)

    DTWBlock dtw_empty_block()
    Py_ssize_t dtw_distances_ptrs(seq_t **ptrs, Py_ssize_t nb_ptrs, Py_ssize_t* lengths, seq_t* output,
                          DTWBlock* block, DTWSettings* settings)
    Py_ssize_t dtw_distances_ndim_ptrs(seq_t **ptrs, Py_ssize_t nb_ptrs, Py_ssize_t* lengths, int ndim, seq_t* output,
                                   DTWBlock* block, DTWSettings* settings)
    Py_ssize_t dtw_distances_matrix(seq_t *matrix, Py_ssize_t nb_rows, Py_ssize_t nb_cols, seq_t* output,
                            DTWBlock* block, DTWSettings* settings)
    Py_ssize_t dtw_distances_ndim_matrix(seq_t *matrix, Py_ssize_t nb_rows, Py_ssize_t nb_cols, int ndim, seq_t* output,
                                     DTWBlock* block, DTWSettings* settings)
    Py_ssize_t dtw_distances_matrices(seq_t *matrix_r, Py_ssize_t nb_rows_r, Py_ssize_t nb_cols_r,
                                      seq_t *matrix_c, Py_ssize_t nb_rows_c, Py_ssize_t nb_cols_c,
                                      seq_t * output, DTWBlock * block, DTWSettings * settings)
    Py_ssize_t dtw_distances_length(DTWBlock *block, Py_ssize_t nb_series_r, Py_ssize_t nb_series_c)

    void dtw_print_wps(seq_t * wps, Py_ssize_t l1, Py_ssize_t l2, DTWSettings * settings)
    void dtw_print_wps_compact(seq_t * wps, Py_ssize_t l1, Py_ssize_t l2, DTWSettings * settings)



# -*- coding: UTF-8 -*-
"""
dtaidistance.dtw
~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW)

:author: Wannes Meert
:copyright: Copyright 2017-2025 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import array
import math

from . import ed
from . import util
from . import util_numpy
from . import innerdistance
from .util import SeriesContainer
from .exceptions import NumpyException, CythonException


logger = logging.getLogger("be.kuleuven.dtai.distance")

dtw_cc = None
try:
    from . import dtw_cc
except ImportError:
    logger.debug('DTAIDistance C library not available')
    dtw_cc = None

dtw_cc_omp = None
dtw_cc_omp_err = None
try:
    from . import dtw_cc_omp
except ImportError as exc:
    dtw_cc_omp_err = str(exc)
    logger.debug('DTAIDistance C-OMP library not available')
    logger.debug(exc)
    dtw_cc_omp = None

dtw_cc_numpy = None
try:
    from . import dtw_cc_numpy
except ImportError:
    logger.debug('DTAIDistance C-Numpy library not available')
    dtw_cc_numpy = None
except ValueError as exc:
    logger.debug('DTAIDistance C-Numpy library not available')
    dtw_cc_numpy = None

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    DTYPE = util_numpy.seq_t_dtype
    argmin = np.argmin
    argmax = np.argmax
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    argmin = util.argmin
    argmax = util.argmax
    array_min = min
    array_max = max


def try_import_c(verbose=False):
    return util.try_import_c(verbose)


inf = float("inf")


def _check_library(include_omp=False, raise_exception=True):
    if dtw_cc is None:
        msg = "The compiled dtaidistance C library is not available.\n" + \
              "See the documentation for alternative installation options."
        logger.error(msg)
        if raise_exception:
            raise CythonException(msg)
    if include_omp and (dtw_cc_omp is None or not dtw_cc_omp.is_openmp_supported()):
        msg = "The compiled dtaidistance C-OMP library "
        if dtw_cc_omp and not dtw_cc_omp.is_openmp_supported():
            msg += "indicates that OpenMP was not avaiable during compilation.\n"
        else:
            msg += "is not available.\n"
        msg += "Use Python's multiprocessing library for parellelization (use_mp=True).\n" + \
               "Call dtw.try_import_c() to get more verbose errors.\n" + \
               "See the documentation for alternative installation options."
        logger.error(msg)
        if raise_exception:
            raise CythonException(msg)


class DTWSettings:
    def __init__(self, window=None, use_pruning=False, max_dist=None, max_step=None,
                 max_length_diff=None, penalty=None, psi=None, inner_dist=innerdistance.default,
                 use_ndim=False, use_c=False):
        """Settings for Dynamic Time Warping distance methods.

        :param window: Only allow for maximal shifts from the two diagonals smaller than this number.
            The maximally allowed warping, thus difference between indices i in series 1 and j in series 2,
            is thus ``|i-j| < window + |len(s1) - len(s2)|``.
            It includes the diagonal, meaning that Euclidean distance is obtained by setting
            ``window=1.`` Setting ``window=0`` or ``window=None`` means that no band is used.
            If the two series are of equal length, this means that the band you see appearing
            on the cumulative cost matrix is of width 2*window-1. In other definitions of DTW
            this width of the band can be referred to as the window.
        :param max_dist: Stop if the returned values will be larger than this value
        :param max_step: Do not allow steps larger than this value.
            If the difference between two values in the two series is larger than this, thus
            if ``|s1[i]-s2[j]| > max_step``, replace that value with infinity.
        :param max_length_diff: Return infinity if length of two series is larger
        :param penalty: Penalty to add if compression or expansion is applied
        :param psi: Psi relaxation parameter (ignore start and end of matching).
            If psi is a single integer, it is used for both start and end relaxations of both series.
            If psi is a 4-tuple, it is used as the psi-relaxation for
            (begin series1, end series1, begin series2, end series2).
            Useful for cyclical series.
        :param use_pruning: Prune values based on Euclidean distance.
            This is the same as passing ub_euclidean() to max_dist
        :param inner_dist: Distance between two points in the time series.
            One of 'squared euclidean' (default), 'euclidean'.
            When using the pure Python implementation (thus use_c=False) then the argument can also
            be an object that has as callable arguments 'inner_dist', 'result', and 'inner_val'. The 'inner_dist'
            function computes the distance between two points (e.g., squared euclidean) and 'result'
            is the function to apply to the final distance (e.g., sqrt when using squared euclidean).
            You can also inherit from the 'innerdistance.CustomInnerDist' class.
        :param use_ndim: Use n-dimensional (aka multivariate) series instead of 1-dimensional series.
        :param use_c: Use the C variant if available.
        """
        self.window = window
        self.use_pruning = use_pruning
        self.max_dist = max_dist
        self.max_step = max_step
        self.max_length_diff = max_length_diff
        self.penalty = penalty
        self.psi = psi
        self.inner_dist = inner_dist
        self.use_ndim = use_ndim
        self.use_c = use_c

        _, _, inner_val = innerdistance.inner_dist_fns(self.inner_dist)

        if not self.max_step:
            self.adj_max_step = inf
        else:
            self.adj_max_step = inner_val(self.max_step)

        if not self.max_dist:
            self.adj_max_dist = inf
        else:
            self.adj_max_dist = inner_val(self.max_dist)

        if not self.penalty:
            self.adj_penalty = 0
        else:
            self.adj_penalty = inner_val(self.penalty)

        if self.max_length_diff is None:
            self.adj_max_length_diff = inf
        else:
            self.adj_max_length_diff = self.max_length_diff

    @staticmethod
    def wrap(settings) -> 'DTWSettings':
        if (isinstance(settings, DTWSettings) or
                'DTWSettings' in str(type(settings))):
            return settings
        if settings is None:
            return DTWSettings()
        if type(settings) is dict and 'dtw_settings' in settings:
            if len(settings) > 1:
                raise ValueError('If dtw_settings is passed, no other arguments'
                                 'can be given.')
            return settings['dtw_settings']
        return DTWSettings(**settings)

    @staticmethod
    def for_dtw(s1, s2, **kwargs) -> 'DTWSettings':
        settings = DTWSettings.wrap(kwargs)
        if settings.window is None:
            settings.window = max(len(s1), len(s2))
        settings.set_max_dist(s1, s2)
        return settings

    def set_max_dist(self, s1, s2):
        _, _, ival_fn = innerdistance.inner_dist_fns(self.inner_dist, use_ndim=self.use_ndim)
        if self.use_pruning and (self.max_dist == 0 or self.max_dist is None):
            self.max_dist = ub_euclidean(s1, s2, inner_dist=self.inner_dist)
            self.adj_max_dist = ival_fn(self.max_dist)

    def kwargs(self):
        return {
            'window': self.window,
            'use_pruning': self.use_pruning,
            'max_dist': self.max_dist,
            'max_step': self.max_step,
            'max_length_diff': self.max_length_diff,
            'penalty': self.penalty,
            'psi': self.psi,
            'inner_dist': self.inner_dist,
            'use_ndim': self.use_ndim,
            'use_c': self.use_c
        }

    def c_kwargs(self):
        window = 0 if self.window is None else self.window
        max_dist = 0 if self.max_dist is None else self.max_dist
        max_step = 0 if self.max_step is None else self.max_step
        max_length_diff = 0 if (self.max_length_diff is None or math.isinf(self.max_length_diff)) \
            else self.max_length_diff
        penalty = 0 if self.penalty is None else self.penalty
        psi = 0 if self.psi is None else self.psi
        use_pruning = 0 if self.use_pruning is None else self.use_pruning
        inner_dist = innerdistance.to_c(self.inner_dist)
        return {
            'window': window,
            'max_dist': max_dist,
            'max_step': max_step,
            'max_length_diff': max_length_diff,
            'penalty': penalty,
            'psi': psi,
            'use_pruning': use_pruning,
            'inner_dist': inner_dist
        }

    def split_psi(self):
        psi_1b = psi_1e = psi_2b = psi_2e = 0
        if type(self.psi) is int:
            psi_1b = psi_1e = psi_2b = psi_2e = self.psi
        elif type(self.psi) in [tuple, list]:
            psi_1b, psi_1e, psi_2b, psi_2e = self.psi
        return psi_1b, psi_1e, psi_2b, psi_2e

    def to_h5_group(self, group):
        for key, value in self.c_kwargs().items():
            group.attrs[key] = value

    @staticmethod
    def from_h5_group(group):
        kwargs = {}
        for attr in ["window", "use_pruning", "max_dist", "max_step",
                     "max_length_diff", "penalty", "psi", "inner_dist", "use_ndim", "use_c"]:
            if attr in group.attrs:
                kwargs[attr] = group.attrs[attr]
        return DTWSettings(**kwargs)

    def __str__(self):
        r = ''
        a = self.c_kwargs()
        for k, v in a.items():
            r += '{}: {}\n'.format(k, v)
        return r


def lb_keogh(s1, s2, **kwargs):
    """Lowerbound LB_KEOGH"""
    s = DTWSettings(**kwargs)
    if s.use_c:
        return dtw_cc.lb_keogh(s1, s2, window=s.window, max_dist=s.max_dist,
                               max_step=s.max_step, inner_dist=s.inner_dist)
    if s.window is None:
        s.window = max(len(s1), len(s2))
    idist_fn, result_fn, _ = innerdistance.inner_dist_fns(s.inner_dist, use_ndim=s.use_ndim)

    t = 0
    imin_diff = max(0, len(s1) - len(s2)) + s.window - 1
    imax_diff = max(0, len(s2) - len(s1)) + s.window
    for i in range(len(s1)):
        imin = max(0, i - imin_diff)
        imax = min(len(s2), i + imax_diff)
        ui = array_max(s2[imin:imax])
        li = array_min(s2[imin:imax])
        ci = s1[i]
        if ci > ui:
            t += idist_fn(ci, ui)  # (ci - ui)**2
        elif ci < li:
            t += idist_fn(ci, li)  # (ci - li)**2
        else:
            pass
    return result_fn(t)


def ub_euclidean(s1, s2, inner_dist=innerdistance.default):
    """ See :meth:`dtaidistance.ed.euclidean_distance`"""
    return ed.distance(s1, s2, inner_dist=inner_dist)


def distance(s1, s2, **kwargs) -> float:
    """
    Dynamic Time Warping.

    This function keeps a compact matrix, not the full warping paths matrix.

    Uses dynamic programming to compute::

        wps[i, j] = (s1[i]-s2[j])**2 + min(
                        wps[i-1, j  ] + penalty,  // vertical   / insertion / expansion
                        wps[i  , j-1] + penalty,  // horizontal / deletion  / compression
                        wps[i-1, j-1])            // diagonal   / match
        dtw = sqrt(wps[-1, -1])

    :param s1: First sequence
    :param s2: Second sequence
    :param kwargs: :class:`DTWSettings` arguments

    Returns: DTW distance
    """
    s = DTWSettings.for_dtw(s1, s2, **kwargs)
    if s.use_c:
        if dtw_cc is None:
            logger.warning("C-library not available, using the Python version")
        else:
            return distance_fast(s1, s2, **s.kwargs())
    idist_fn, result_fn, ival_fn = innerdistance.inner_dist_fns(s.inner_dist, use_ndim=s.use_ndim)
    r, c = len(s1), len(s2)
    if s.adj_max_length_diff is not None and abs(r - c) > s.adj_max_length_diff:
        return inf
    window = max(len(s1), len(s2)) if s.window is None else s.window

    psi_1b, psi_1e, psi_2b, psi_2e = s.split_psi()
    length = min(c + 1, abs(r - c) + 2 * (window - 1) + 1 + 1 + 1)
    dtw = array.array('d', [inf] * (2 * length))
    sc = 0
    ec = 0
    for i in range(psi_2b + 1):
        dtw[i] = 0
    skip = 0
    i0 = 1
    i1 = 0
    psi_shortest = inf
    for i in range(r):
        # print("i={}".format(i))
        # print(dtw)
        skipp = skip
        skip = max(0, i - max(0, r - c) - window + 1)
        i0 = 1 - i0
        i1 = 1 - i1
        for ii in range(i1*length, i1*length+length):
            dtw[ii] = inf
        j_start = max(0, i - max(0, r - c) - window + 1)
        j_end = min(c, i + max(0, c - r) + window)
        if sc > j_start:
            j_start = sc
        smaller_found = False
        ec_next = i
        if length == c + 1:
            skip = 0
        if psi_1b != 0 and j_start == 0 and i < psi_1b:
            dtw[i1 * length] = 0
        for j in range(j_start, j_end):
            # d = (s1[i] - s2[j])**2
            d = idist_fn(s1[i], s2[j])
            if d > s.adj_max_step:
                continue
            assert j + 1 - skip >= 0
            assert j - skipp >= 0
            assert j + 1 - skipp >= 0
            assert j - skip >= 0
            dtw[i1 * length + j + 1 - skip] = d + min(dtw[i0 * length + j - skipp],
                                                      dtw[i0 * length + j + 1 - skipp] + s.adj_penalty,
                                                      dtw[i1 * length + j - skip] + s.adj_penalty)
            # print('({},{}), ({},{}), ({},{})'.format(i0, j - skipp, i0, j + 1 - skipp, i1, j - skip))
            # print('{}, {}, {}'.format(dtw[i0, j - skipp], dtw[i0, j + 1 - skipp], dtw[i1, j - skip]))
            # print('i={}, j={}, d={}, skip={}, skipp={}'.format(i,j,d,skip,skipp))
            # print(dtw)
            if dtw[i1 * length + j + 1 - skip] > s.adj_max_dist:
                if not smaller_found:
                    sc = j + 1
                if j >= ec:
                    break
            else:
                smaller_found = True
                ec_next = j + 1
        ec = ec_next
        if psi_1e != 0 and j_end == len(s2) and len(s1) - 1 - i <= psi_1e:
            psi_shortest = min(psi_shortest, dtw[i1 * length + j_end - skip])
    if psi_1e == 0 and psi_2e == 0:
        d = dtw[i1 * length + min(c, c + window - 1) - skip]
    else:
        ic = min(c, c + window - 1) - skip
        if psi_2e != 0:
            vc = dtw[i1 * length + ic - psi_2e:i1 * length + ic + 1]
            d = min(array_min(vc), psi_shortest)
        else:
            d = min(dtw[i1 * length + min(c, c + window - 1) - skip], psi_shortest)
    if s.adj_max_dist and d > s.adj_max_dist:
        d = inf
    d = result_fn(d)
    return d


def distance_fast(s1, s2, use_pruning=True, **kwargs):
    """Same as :meth:`distance` but with different defaults to choose the fast C-based version of
    the implementation (use_c = True) and use pruning (use_pruning = True).

    Note: the series are expected to be arrays of the type ``double``.
    Thus ``numpy.array([1,2,3], dtype=numpy.double)`` or
    ``array.array('d', [1,2,3])``
    """
    _check_library(raise_exception=True)
    # Check that Numpy arrays for C contiguous
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    s = DTWSettings(use_pruning=use_pruning, **kwargs)
    # Move data to C library
    if s.use_ndim is False:
        d = dtw_cc.distance(s1, s2, **s.c_kwargs())
    else:
        d = dtw_cc.distance_ndim(s1, s2, **s.c_kwargs())
    return d


def _distance_with_params(t):
    return distance(t[0], t[1], **t[2])


def _distance_with_params_ndim(t):
    return distance(t[0], t[1], use_ndim=True, **t[2])


def _distance_c_with_params(t):
    return dtw_cc.distance(t[0], t[1], **t[2])


def _distance_c_with_params_ndim(t):
    return dtw_cc.distance_ndim(t[0], t[1], **t[2])


def warping_paths(s1, s2, psi_neg=True, keep_int_repr=False, **kwargs):
    """
    Dynamic Time Warping.

    The full matrix of all warping paths (or accumulated cost matrix) is built.

    :param s1: First sequence
    :param s2: Second sequence
    :param psi_neg: Replace values that should be skipped because of psi-relaxation with -1.
    :param keep_int_repr: Keep the internal representation of the numbers.
        For example, by default this is the squared differences.
    :param kwargs: See arguments for :class:`DTWSettings`
    :returns: (DTW distance, DTW matrix)
    """
    s = DTWSettings.for_dtw(s1, s2, **kwargs)
    if s.use_c:
        return warping_paths_fast(s1, s2, psi_neg=psi_neg, **s.kwargs())
    if np is None:
        raise NumpyException("Numpy is required for the warping_paths method")
    cost, result_fn, _ = innerdistance.inner_dist_fns(s.inner_dist, use_ndim=s.use_ndim)
    r, c = len(s1), len(s2)
    if s.adj_max_length_diff is not None and abs(r - c) > s.adj_max_length_diff:
        return inf, None
    window = max(len(s1), len(s2)) if s.window is None else s.window
    psi_1b, psi_1e, psi_2b, psi_2e = s.split_psi()
    dtw = np.full((r + 1, c + 1), inf)
    # dtw[0, 0] = 0
    for i in range(psi_2b + 1):
        dtw[0, i] = 0
    for i in range(psi_1b + 1):
        dtw[i, 0] = 0
    i1 = 0
    sc = 0
    ec = 0
    for i in range(r):
        i0 = i
        i1 = i + 1
        j_start = max(0, i - max(0, r - c) - window + 1)
        j_end = min(c, i + max(0, c - r) + window)
        if sc > j_start:
            j_start = sc
        smaller_found = False
        ec_next = i
        for j in range(j_start, j_end):
            d = cost(s1[i], s2[j])
            if s.adj_max_step is not None and d > s.adj_max_step:
                continue
            dtw[i1, j + 1] = d + min(dtw[i0, j],
                                     dtw[i0, j + 1] + s.adj_penalty,
                                     dtw[i1, j] + s.adj_penalty)
            if dtw[i1, j + 1] > s.adj_max_dist:
                if not smaller_found:
                    sc = j + 1
                if j >= ec:
                    break
            else:
                smaller_found = True
                ec_next = j + 1
        ec = ec_next
    # Decide which d to return
    if not keep_int_repr:
        dtw = result_fn(dtw)
    if psi_1e == 0 and psi_2e == 0:
        d = dtw[i1, min(c, c + window - 1)]
    else:
        ir = i1
        ic = min(c, c + window - 1)
        if psi_1e != 0:
            vr = dtw[ir:max(0, ir-psi_1e-1):-1, ic]
            mir = argmin(vr)
            vr_mir = vr[mir]
        else:
            mir = ir
            vr_mir = inf
        if psi_2e != 0:
            vc = dtw[ir, ic:max(0, ic-psi_2e-1):-1]
            mic = argmin(vc)
            vc_mic = vc[mic]
        else:
            mic = ic
            vc_mic = inf
        if vr_mir < vc_mic:
            if psi_neg:
                dtw[ir:ir-mir:-1, ic] = -1
            d = vr_mir
        else:
            if psi_neg:
                dtw[ir, ic:ic-mic:-1] = -1
            d = vc_mic
    if keep_int_repr:
        if s.adj_max_dist and d > s.adj_max_dist:
            d = inf
    else:
        if s.max_dist and d > s.max_dist:
            d = inf
    return d, dtw


def warping_paths_fast(s1, s2, psi_neg=True, keep_int_repr=False, compact=False, **kwargs):
    """Fast C version of :meth:`warping_paths`.

    The `use_pruning` argument is still False by default in case one needs the
    full matrix. If this is not the case, then `use_pruning` can be set to True.

    :param s1: See :meth:`warping_paths`
    :param s2: See :meth:`warping_paths`
    :param psi_neg: See :meth:`warping_paths`
    :param compact: Return a compact warping paths matrix.
        Size is ((l1 + 1), min(l2 + 1, abs(l1 - l2) + 2*window + 1)).
        This option is meant for internal use. For more details, see the C code.
    :param kwargs: See :meth:`warping_paths`
    """
    if np is None:
        raise util_numpy.NumpyException("Numpy needed for warping_paths_fast")
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    r = len(s1)
    c = len(s2)
    _check_library(raise_exception=True)
    settings = DTWSettings.for_dtw(s1, s2, **kwargs)
    if compact:
        wps_width = dtw_cc.wps_width(r, c, **settings.c_kwargs())
        wps_compact = np.full((len(s1)+1, wps_width), inf)
        if settings.use_ndim:
            d = dtw_cc.warping_paths_compact_ndim(wps_compact, s1, s2, psi_neg, keep_int_repr, **settings.c_kwargs())
        else:
            d = dtw_cc.warping_paths_compact(wps_compact, s1, s2, psi_neg, keep_int_repr, **settings.c_kwargs())
        return d, wps_compact

    dtw = np.full((r + 1, c + 1), inf)
    if settings.use_ndim:
        d = dtw_cc.warping_paths_ndim(dtw, s1, s2, psi_neg, keep_int_repr, **settings.c_kwargs())
    else:
        d = dtw_cc.warping_paths(dtw, s1, s2, psi_neg, keep_int_repr, **settings.c_kwargs())
    return d, dtw


def warping_paths_affinity(s1, s2, window=None, only_triu=False,
                           penalty=None, psi=None, psi_neg=True,
                           gamma=1, tau=0, delta=0, delta_factor=1,
                           use_c=False):
    """
    Dynamic Time Warping paths using an affinity/similarity matrix instead of a distance matrix.

    The full matrix of all warping paths (or accumulated cost matrix) is built.

    :param s1: First sequence
    :param s2: Second sequence
    :param window: see :meth:`distance`
    :param only_triu: Only compute upper traingular matrix of warping paths.
        This is useful if s1 and s2 are the same series and the matrix would be mirrored around the diagonal.
    :param penalty: see :meth:`distance`
    :param psi: see :meth:`distance`
    :param psi_neg: Replace values that should be skipped because of psi-relaxation with -1.
    :param gamma:
    :param tau:
    :param delta:
    :param delta_factor:
    :param use_c:
    :returns: (DTW distance, DTW matrix)
    """
    if use_c:
        return warping_paths_affinity_fast(s1, s2, window=window, only_triu=only_triu,
                                           penalty=penalty, tau=tau, delta=delta, delta_factor=delta_factor)
    if np is None:
        raise NumpyException("Numpy is required for the warping_paths method")
    s = DTWSettings.for_dtw(s1, s2, window=window, psi=psi, penalty=penalty, use_c=use_c)
    r, c = len(s1), len(s2)
    psi_1b, psi_1e, psi_2b, psi_2e = s.split_psi()
    window = max(len(s1), len(s2)) if s.window is None else s.window
    dtw = np.full((r + 1, c + 1), -inf)
    # dtw[0, 0] = 0
    for i in range(psi_2b + 1):
        dtw[0, i] = 0
    for i in range(psi_1b + 1):
        dtw[i, 0] = 0
    i1 = 0
    for i in range(r):
        i0 = i
        i1 = i + 1
        j_start = max(0, i - max(0, r - c) - window + 1)
        if only_triu:
            j_start = max(i, j_start)
        j_end = min(c, i + max(0, c - r) + window)
        for j in range(j_start, j_end):
            d = np.exp(-gamma*(s1[i] - s2[j])**2)
            # print(f"{s1[i] - s2[j]=} -> {d=}")
            dtw_prev = max(dtw[i0, j],
                           dtw[i0, j + 1] - s.penalty,
                           dtw[i1, j] - s.penalty)
            if d < tau:
                dtw[i1, j + 1] = max(0, delta + delta_factor * dtw_prev)
            else:
                dtw[i1, j + 1] = max(0, d + dtw_prev)

    # Decide which d to return
    if psi_1e == 0 and psi_2e == 0:
        d = dtw[i1, min(c, c + window - 1)]
    else:
        ir = i1
        ic = min(c, c + window - 1)
        if psi_1e != 0:
            vr = dtw[ir:max(0, ir-psi_1e-1):-1, ic]
            mir = argmax(vr)
            vr_mir = vr[mir]
        else:
            mir = ir
            vr_mir = inf
        if psi_2e != 0:
            vc = dtw[ir, ic:max(0, ic-psi_2e-1):-1]
            mic = argmax(vc)
            vc_mic = vc[mic]
        else:
            mic = ic
            vc_mic = inf
        if vr_mir > vc_mic:
            if psi_neg:
                dtw[ir:ir-mir:-1, ic] = -1
            d = vr_mir
        else:
            if psi_neg:
                dtw[ir, ic:ic-mic:-1] = -1
            d = vc_mic
    return d, dtw


def warping_paths_affinity_fast(s1, s2, window=None, only_triu=False,
                                penalty=None, psi=None, psi_neg=True,
                                gamma=1, tau=0, delta=0, delta_factor=1,
                                compact=False, use_ndim=False):
    """Fast C version of :meth:`warping_paths`.

    :param s1:
    :param s2:
    :param window:
    :param only_triu:
    :param penalty:
    :param psi:
    :param psi_neg:
    :param gamma:
    :param tau:
    :param delta:
    :param delta_factor:
    :param compact: Return a compact warping paths matrix.
        Size is ((l1 + 1), min(l2 + 1, abs(l1 - l2) + 2*window + 1)).
        This option is meant for internal use. For more details, see the C code.
    :param use_ndim:
    """
    if np is None:
        raise util_numpy.NumpyException("Numpy needed for warping_paths_affinity_fast")
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    r = len(s1)
    c = len(s2)
    _check_library(raise_exception=True)
    settings = DTWSettings.for_dtw(s1, s2, window=window, penalty=penalty, psi=psi)
    if compact:
        wps_width = dtw_cc.wps_width(r, c, **settings.c_kwargs())
        wps_compact = np.full((len(s1)+1, wps_width), -inf)
        if use_ndim:
            d = dtw_cc.warping_paths_compact_ndim_affinity(wps_compact, s1, s2, only_triu,
                                                           gamma, tau, delta, delta_factor, psi_neg,
                                                           **settings.c_kwargs())
        else:
            d = dtw_cc.warping_paths_compact_affinity(wps_compact, s1, s2, only_triu,
                                                      gamma, tau, delta, delta_factor, psi_neg, **settings.c_kwargs())
        return d, wps_compact

    dtw = np.full((r + 1, c + 1), -inf)
    if use_ndim:
        d = dtw_cc.warping_paths_affinity_ndim(dtw, s1, s2, only_triu,
                                               gamma, tau, delta, delta_factor, psi_neg, **settings.c_kwargs())
    else:
        d = dtw_cc.warping_paths_affinity(dtw, s1, s2, only_triu,
                                          gamma, tau, delta, delta_factor, psi_neg, **settings.c_kwargs())
    return d, dtw


def distance_matrix_func(use_c=False, parallel=False, show_progress=False):
    def distance_matrix_wrapper(seqs, **kwargs):
        return distance_matrix(seqs, parallel=parallel, use_c=use_c,
                               show_progress=show_progress, **kwargs)
    return distance_matrix_wrapper


def distance_matrix(s, block=None, compact=False, parallel=False,
                    use_mp=False, show_progress=False, only_triu=False, **kwargs):
    """Distance matrix for all sequences in s.

    :param s: Iterable of series
    :param block: Only compute block in matrix. Expects tuple with begin and end, e.g. ((0,10),(20,25)) will
        only compare rows 0:10 with rows 20:25.
    :param compact: Return the distance matrix as an array representing the upper triangular matrix.
    :param parallel: Use parallel operations
    :param use_mp: Force use Multiprocessing for parallel operations (not OpenMP)
    :param show_progress: Show progress using the tqdm library. This is only supported for
        the pure Python version (thus not the C-based implementations).
    :param only_triu: Only compute upper traingular matrix of warping paths.
        This is useful if s1 and s2 are the same series and the matrix would be mirrored around the diagonal.
    :param kwargs: See arguments for :class:`DTWSettings`
    :returns: The distance matrix or the condensed distance matrix if the compact argument is true
    """
    settings = DTWSettings(**kwargs)
    # Check whether multiprocessing is available
    if settings.use_c:
        requires_omp = parallel and not use_mp
        _check_library(raise_exception=True, include_omp=requires_omp)
    if parallel and (use_mp or not settings.use_c):
        try:
            import multiprocessing as mp
            logger.info('Using multiprocessing')
        except ImportError:
            msg = 'Cannot load multiprocessing'
            logger.error(msg)
            raise Exception(msg)
    else:
        mp = None
    if block is not None:
        if len(block) > 2 and block[2] is False and compact is False:
            raise Exception(f'Block cannot have a third argument triu=false with compact=false')
        if (block[0][1] - block[0][0]) < 1 or (block[1][1] - block[1][0]) < 1:
            return []
    # Prepare options and data to pass to distance method
    dist_opts = settings.kwargs()
    s = SeriesContainer.wrap(s)
    if settings.use_ndim:
        ndim = s.detected_ndim
    else:
        ndim = 1
    if settings.use_c:
        for k, v in dist_opts.items():
            if v is None:
                # None is represented as 0.0 for C
                dist_opts[k] = 0

    logger.info('Computing distances')
    if settings.use_c and parallel and not use_mp and dtw_cc_omp is not None:
        logger.info("Compute distances in C (parallel=OMP)")
        dist_opts['block'] = block
        if settings.use_ndim:
            dists = dtw_cc_omp.distance_matrix_ndim(s, ndim, **dist_opts)
        else:
            dists = dtw_cc_omp.distance_matrix(s, **dist_opts)

    elif settings.use_c and parallel and (dtw_cc_omp is None or use_mp):
        logger.info("Compute distances in C (parallel=MP)")
        idxs = _distance_matrix_idxs(block, len(s))
        if settings.use_ndim:
            fn = _distance_c_with_params_ndim
        else:
            fn = _distance_c_with_params
        with mp.Pool() as p:
            dists = p.map(fn, [(s[r], s[c], dist_opts) for c, r in zip(*idxs)])

    elif settings.use_c and not parallel:
        logger.info("Compute distances in C (parallel=No)")
        dist_opts['block'] = block
        if settings.use_ndim:
            dists = dtw_cc.distance_matrix_ndim(s, ndim, **dist_opts)
        else:
            dists = dtw_cc.distance_matrix(s, **dist_opts)

    elif not settings.use_c and parallel:
        logger.info("Compute distances in Python (parallel=MP)")
        idxs = _distance_matrix_idxs(block, len(s))
        if settings.use_ndim:
            fn = _distance_with_params_ndim
        else:
            fn = _distance_with_params
        with mp.Pool() as p:
            dists = p.map(fn, [(s[r], s[c], dist_opts) for c, r in zip(*idxs)])

    elif not settings.use_c and not parallel:
        logger.info("Compute distances in Python (parallel=No)")
        dists = distance_matrix_python(s, block=block, show_progress=show_progress, settings=settings)

    else:
        raise Exception(f'Unsupported combination of: parallel={parallel}, '
                        f'use_c={s.use_c}, dtw_cc_omp={dtw_cc_omp}, use_mp={use_mp}')

    exp_length = _distance_matrix_length(block, len(s))
    assert len(dists) == exp_length, "len(dists)={} != {} (block = {})".format(len(dists), exp_length, block)
    if compact:
        return dists

    # Create full matrix and fill upper triangular matrix with distance values (or only block if specified)
    dists_matrix = distances_array_to_matrix(dists, nb_series=len(s), block=block, only_triu=only_triu)

    return dists_matrix


def distances_array_to_matrix(dists, nb_series, block=None, only_triu=False):
    """Transform a condensed distances array to a full matrix representation.

    The upper triangular matrix will contain all the distances.
    """
    if np is None:
        raise NumpyException("Numpy is required for the distances_array_to_matrix method, "
                             "set compact to true")
    dists_matrix = np.full((nb_series, nb_series), inf, dtype=DTYPE)
    idxs = _distance_matrix_idxs(block, nb_series)
    dists_matrix[idxs] = dists
    if not only_triu:
        dists_matrix.T[idxs] = dists
        np.fill_diagonal(dists_matrix, 0)
    # dists_cond = np.zeros(self._size_cond(len(series)))
    # idx = 0
    # for r in range(len(series) - 1):
    #     dists_cond[idx:idx + len(series) - r - 1] = dists[r, r + 1:]
    #     idx += len(series) - r - 1
    return dists_matrix


def distance_array_index(a, b, nb_series):
    if a == b:
        raise ValueError("Distance between the same series is not available.")
    if a > b:
        a, b = b, a
    idx = 0
    for r in range(a):
        idx += nb_series - r - 1
    idx += b - a - 1
    return idx


def distance_matrix_python(s, block=None, show_progress=False, settings=None):
    if settings is None:
        settings = DTWSettings()
    dists = array.array('d', [inf] * _distance_matrix_length(block, len(s)))
    block, triu = _complete_block(block, len(s))
    it_r = range(block[0][0], block[0][1])
    if show_progress:
        if tqdm is None:
            raise ValueError('show_progress cannot be true is tqdm is not available')
        it_r = tqdm(it_r)
    idx = 0
    for r in it_r:
        if triu:
            it_c = range(max(r + 1, block[1][0]), min(len(s), block[1][1]))
        else:
            it_c = range(block[1][0], min(len(s), block[1][1]))
        for c in it_c:
            dists[idx] = distance(s[r], s[c], **settings.kwargs())
            idx += 1
    return dists


def _complete_block(block, nb_series):
    """Expand block variable to represent exact indices of ranges.

    :param block: None, 0, or tuple
    :param nb_series: Number of series in the list
    :return: Block with filled in indices, Boolean to indicate triu
    """
    if block is None or block == 0:
        block = ((0, nb_series), (0, nb_series))
        return block, True
    else:
        if len(block) > 2 and block[2] is False:
            return block, False
        else:
            return block, True


def _distance_matrix_idxs(block, nb_series):
    if block is None or block == 0:
        if np is not None:
            idxs = np.triu_indices(nb_series, k=1)
            return idxs
    # Numpy is not available or not triu
    block, triu = _complete_block(block, nb_series)
    idxsl_r = []
    idxsl_c = []
    for r in range(block[0][0], block[0][1]):
        if triu:
            it_c = range(max(r + 1, block[1][0]), min(nb_series, block[1][1]))
        else:
            it_c = range(block[1][0], min(nb_series, block[1][1]))
        for c in it_c:
            idxsl_r.append(r)
            idxsl_c.append(c)
    if np is not None:
        idxs = (np.array(idxsl_r), np.array(idxsl_c))
    else:
        idxs = (idxsl_r, idxsl_c)
    return idxs


def _distance_matrix_length(block, nb_series):
    if block is not None:
        block_rb = block[0][0]
        block_re = block[0][1]
        block_cb = block[1][0]
        block_ce = block[1][1]
        length = 0
        if len(block) > 2 and block[2] is False:
            length = (block_re - block_rb) * (block_ce - block_cb)
        else:
            for ri in range(block_rb, block_re):
                if block_cb <= ri:
                    if block_ce > ri:
                        length += (block_ce - ri - 1)
                else:
                    if block_ce > ri:
                        length += (block_ce - block_cb)
    else:
        length = int(nb_series * (nb_series - 1) / 2)
    return length


def distance_matrix_fast(s, max_dist=None, use_pruning=True, max_length_diff=None,
                         window=None, max_step=None, penalty=None, psi=None,
                         block=None, compact=False, parallel=True, use_mp=False,
                         only_triu=False, inner_dist=innerdistance.default, use_c=True):
    """Same as :meth:`distance_matrix` but with different defaults to choose the
    fast parallized C version (use_c = True, parallel = True, use_pruning = True).

    This method uses the C-compiled version of the DTW algorithm and uses parallelization.
    By default, this is the OMP C parallelization. If the OMP functionality is not available
    the parallelization is changed to use Python's multiprocessing library.
    """
    _check_library(raise_exception=True, include_omp=False)
    if not use_mp and parallel:
        try:
            _check_library(raise_exception=True, include_omp=True)
        except CythonException:
            use_mp = True
    return distance_matrix(s, max_dist=max_dist, use_pruning=use_pruning,
                           max_length_diff=max_length_diff, window=window,
                           max_step=max_step, penalty=penalty, psi=psi,
                           block=block, compact=compact, parallel=parallel,
                           use_c=True, use_mp=use_mp, show_progress=False,
                           only_triu=only_triu, inner_dist=inner_dist)


def warping_path(from_s, to_s, include_distance=False, use_ndim=False, **kwargs):
    """Compute the warping path between two sequences.

    :param from_s: First time series
    :param to_s: Second time series
    :param include_distance: Change the return values to a tuple with
        (list of coordinates, distance value)
    :param use_ndim: The data is multi-dimensional (or multi-variate)
    :param \\**kwargs: Additional options that are passed to :class:`DTWSettings`
    :return: List of 2d path coordinates
    """
    dist, paths = warping_paths(from_s, to_s, use_ndim=use_ndim, **kwargs)
    path = best_path(paths)
    if include_distance:
        return path, dist
    return path


def warping_path_fast(from_s, to_s, include_distance=False, use_lowmem=False,
                      **kwargs):
    """Compute warping path between two sequences.

    :param use_lowmem: Switch to a low-memory version of the warping path
        algorithm based on the Hirschberg algorithm (at the expense of speed).
        If, instead of a boolean, an int value is given then this will be
        used to set the largest block that can be computed at maximal
        memory usage (i.e, O(value*value)).
        This option is not compatible with the Psi-relaxiation.
    """
    if len(from_s.shape) > 1:
        ndim = from_s.shape[1]
    else:
        ndim = 1
    switch_to_full = 1000
    if type(use_lowmem) is int:
        switch_to_full = use_lowmem
    from_s, to_s, settings_kwargs = warping_path_args_to_c(from_s, to_s, **kwargs)
    if use_lowmem:
        if "psi" in kwargs:
            raise ValueError("The argument psi is not supported when use_lowmem=True")
        if "max_step" in kwargs:
            raise ValueError("The argument max_step is not supported when use_lowmem=True")
        if "max_length_diff" in kwargs:
            raise ValueError("The argument max_length_diff is not supported when use_lowmem=True")
        if ndim == 1:
            path, d = dtw_cc.warping_path_lowmem(
                from_s, to_s, switch_to_full=switch_to_full,
                ndim=ndim, **settings_kwargs)
        else:
            path, d = dtw_cc.warping_path_lowmem_ndim(
                from_s, to_s, switch_to_full=switch_to_full,
                ndim=ndim, **settings_kwargs)
        if include_distance:
            result = path, d
        result = path
    elif ndim > 1:
        result = dtw_cc.warping_path_ndim(
            from_s, to_s, ndim=ndim, include_distance=include_distance,
            **settings_kwargs)
    else:
        result = dtw_cc.warping_path(
            from_s, to_s, include_distance=include_distance,
            **settings_kwargs)
    return result


def warping_path_prob(from_s, to_s, avg, include_distance=False, use_c=True, **kwargs):
    """Compute warping path between two sequences."""
    if not use_c:
        raise AttributeError('warping_path_prob with use_c=False not yet supported')
    from_s, to_s, settings_kwargs = warping_path_args_to_c(from_s, to_s, **kwargs)
    result = dtw_cc.warping_path_prob(from_s, to_s, avg,
                                      include_distance=include_distance, **settings_kwargs)
    return result


def warping_amount(path):
    """
        Returns the number of compressions and expansions performed to obtain the best path.
        Can be used as a metric for the amount of warping.

        :param path: path to be tested

        :returns number of compressions or expansions

    """
    n = 0
    for i in range(1, len(path)):
        if path[i - 1][0] + 1 != path[i][0] or path[i - 1][1] + 1 != path[i][1]:
            n += 1

    return n


def warping_path_penalty(s1, s2, penalty_post=0, **kwargs):
    """Dynamic Time Warping with an alternative penalty.

    This function supports two different penalties. The traditional DTW penalty `penalty`
    is used in the matrix during calculation of the warping path (see :meth:`distance`).

    The second penalty `penalty_post` measures the amount of warping. This penalty doesn't
    affect the warping path and is added to the DTW distance after the warping for every compression or expansion.

    Same options as :meth:`warping_paths`

    :param s1: First sequence
    :param s2: Second sequence
    :param penalty_post: Penalty to be added after path calculation, for compression/extension

    :returns DTW distance, the best path, DTW distance between 2 path elements, DTW matrix
    """
    dist, paths = warping_paths(s1, s2, **kwargs)
    path = best_path(paths)

    path_stepsize = []
    for i in range(1, len(path)):
        if path[i - 1][0] + 1 != path[i][0] or path[i - 1][1] + 1 != path[i][1]:
            dist += penalty_post

        path_stepsize.append(paths[path[i][0] + 1, path[i][1] + 1] - paths[path[i - 1][0] + 1, path[i - 1][1] + 1])

    return [dist, path, path_stepsize, paths]


def warp(from_s, to_s, path=None, **kwargs):
    """Warp a function to optimally match a second function.

    :param from_s: First sequence
    :param to_s: Second sequence
    :param path: (Optional) Path to use wrap the 'from_s' sequence to the 'to_s' sequence
                If provided, this function will use it.
                If not provided, this function will calculate it using the warping_path function
    :param kwargs: Same options as :meth:`warping_paths`.
    """
    if path is None:
        path = warping_path(from_s, to_s, **kwargs)
    from_s2 = array.array('d', [0] * len(to_s))
    from_s2_cnt = array.array('i', [0] * len(to_s))
    for r_c, c_c in path:
        from_s2[c_c] += from_s[r_c]
        from_s2_cnt[c_c] += 1
    for i in range(len(to_s)):
        from_s2[i] /= from_s2_cnt[i]
    return from_s2, path


def best_path(paths, row=None, col=None, use_max=False, penalty=0):
    """Compute the optimal path from the nxm warping paths matrix.

    :param paths: Warping paths matrix
    :param row: If given, start from this row (instead of lower-right corner)
    :param col: If given, start from this column (instead of lower-right corner)
    :param use_max: Find maximal path instead of minimal path
    :param penalty: The penalty used.
        If this is used, then paths should be expressed as the internal representation.
    :return: Array of (row, col) or (ts_1_idx, ts_2_idx) representing the best path
    """
    if use_max:
        argm = argmax
    else:
        argm = argmin
    if row is None:
        i = int(paths.shape[0] - 1)
    else:
        i = row
    if col is None:
        j = int(paths.shape[1] - 1)
    else:
        j = col
    p = []
    if paths[i, j] != -1:
        p.append((i - 1, j - 1))
    while i > 0 and j > 0:
        c = argm([paths[i - 1, j - 1],
                  paths[i - 1, j] + penalty,
                  paths[i, j - 1] + penalty])
        if c == 0:
            i, j = i - 1, j - 1
        elif c == 1:
            i = i - 1
        elif c == 2:
            j = j - 1
        if paths[i, j] != -1:
            p.append((i - 1, j - 1))
    p.pop()
    p.reverse()
    return p


def best_path2(paths):
    """Compute the optimal path from the nxm warping paths matrix."""
    if np is None:
        raise util_numpy.NumpyException("Numpy needed for best_path2")
    m = paths
    path = []
    r, c = m.shape
    r -= 1
    c -= 1
    v = m[r, c]
    if v != -1:
        path.append((r - 1, c - 1))
    while r > 0 and c > 0:
        if v == -1:
            v = np.inf
        r_c, c_c = r, c
        if r >= 1 and c >= 1 and m[r - 1, c - 1] <= v:
            r_c, c_c, v = r - 1, c - 1, m[r - 1, c - 1]
        if r >= 1 and m[r - 1, c] <= v:
            r_c, c_c, v = r - 1, c, m[r - 1, c]
        if c >= 1 and m[r, c - 1] <= v:
            r_c, c_c, v = r, c - 1, m[r, c - 1]
        if v != -1:
            path.append((r_c - 1, c_c - 1))
        r, c = r_c, c_c
    path.pop()
    path.reverse()
    return path


def warping_path_args_to_c(s1, s2, **kwargs):
    s1 = util_numpy.verify_np_array(s1)
    s2 = util_numpy.verify_np_array(s2)
    _check_library(raise_exception=True)

    def get(key):
        value = kwargs.get(key, None)
        if value is None:
            return 0
        return value
    settings_kwargs = {key: get(key) for key in
                       ['window', 'max_dist', 'max_step', 'max_length_diff', 'penalty', 'psi']}
    return s1, s2, settings_kwargs




"""
dtaidistance.loco_cc
~~~~~~~~~~~~~~~~~~~~

Local Concurrences, C implementation.

:author: Wannes Meert
:copyright: Copyright 2023-2025 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
cimport dtaidistancec_loco
cimport dtaidistancec_globals
from dtaidistancec_dtw cimport seq_t
from cython.view cimport array as cvarray
import array
from libc.stdio cimport FILE, fopen, fwrite, fscanf, fclose, fprintf, fseek, ftell, SEEK_END, rewind, fread

logger = logging.getLogger("be.kuleuven.dtai.distance")


def save_wp(seq_t[:, :] wps, filename):
    cdef bytes py_bytes = filename.encode()
    cdef char * c_string = py_bytes
    cdef Py_ssize_t N = wps.shape[0] * wps.shape[1]
    cdef FILE *ptr_fw
    cdef seq_t *ptr_d = &wps[0,0]
    ptr_fw = fopen(c_string, "wb")
    fwrite(ptr_d, sizeof(seq_t), N, ptr_fw)
    fclose(ptr_fw)


cdef class LoCoSettings:
    def __cinit__(self):
        pass

    def __init__(self, **kwargs):
        self._settings = dtaidistancec_loco.loco_settings_default()

        self._settings.window = 0
        if "window" in kwargs:
            if kwargs["window"] is None:
                pass
            else:
                self._settings.window = kwargs["window"]

        self._settings.penalty = 0
        if "penalty" in kwargs:
            if kwargs["penalty"] is None:
                pass
            else:
                self._settings.penalty = kwargs["penalty"]

        self._settings.psi_1b = self._settings.psi_2b = 0
        if "psi" in kwargs:
            if kwargs["psi"] is None:
                pass
            elif type(kwargs["psi"]) is int:
                self._settings.psi_1b = self._settings.psi_2b = kwargs["psi"]
            elif type(kwargs["psi"]) is tuple or type(kwargs["psi"]) is list:
                self._settings.psi_1b, self._settings.psi_2b = kwargs["psi"]

        self._settings.only_triu = kwargs.get("only_triu", 0)
        self._settings.gamma = kwargs.get("gamma", 0)
        self._settings.tau = kwargs.get("tau", 0)
        self._settings.delta = kwargs.get("delta", 0)
        self._settings.delta_factor = kwargs.get("delta_factor", 0)

        if "step_type" in kwargs:
            step_type = kwargs["step_type"]
            if step_type == "TypeI":
                self._settings.step_type = dtaidistancec_globals.TypeI
            elif step_type == "TypeIII":
                self._settings.step_type = dtaidistancec_globals.TypeIII
            else:
                raise AttributeError("Unknown step_type: {}".format(step_type))

    @property
    def step_type(self):
        return self._settings.step_type

    def inf_rowscols(self):
        if self._settings.step_type == dtaidistancec_globals.TypeI:
            return 1, 1
        elif self._settings.step_type == dtaidistancec_globals.TypeIII:
            return 2, 2
        else:
            raise AttributeError("Unknown step_type: {}".format(self._settings.step_type))


def loco_warping_paths(seq_t[:, :] wps, seq_t[:] s1, seq_t[:] s2, ndim=1, **kwargs):
    settings = LoCoSettings(**kwargs)
    if settings.step_type == dtaidistancec_globals.TypeI:
        settings._settings.step_type = dtaidistancec_globals.TypeI
        dtaidistancec_loco.loco_warping_paths_ndim_typeI(&wps[0,0], &s1[0], len(s1), &s2[0], len(s2),
                                                         ndim, &settings._settings)
    elif settings.step_type == dtaidistancec_globals.TypeIII:
        settings._settings.step_type = dtaidistancec_globals.TypeIII
        dtaidistancec_loco.loco_warping_paths_ndim_typeIII(&wps[0,0], &s1[0], len(s1), &s2[0], len(s2),
                                                           ndim, &settings._settings)
    else:
        raise ValueError("Unknown steps type.")

def loco_best_path(seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t r, Py_ssize_t c, int min_size, **kwargs):
    settings = LoCoSettings(**kwargs)
    inf_rows, inf_cols = settings.inf_rowscols()
    # print(f"{inf_rows=}, {inf_cols=}, {r=}, {c=}")
    cdef Py_ssize_t width = l2 + inf_cols
    cdef dtaidistancec_globals.DDPath bp = dtaidistancec_loco.loco_best_path(&wps[0,0], l1, l2, r, c, min_size,
                                                                             &settings._settings)
    if bp.length == 0:
        print('ERROR bp.length = 0')
        return None
    try:
        # Use cython.view.array to avoid numpy dependency
        # idxs_a = array.array('q', [0] * (bp.length * 2))
        # idxs = memoryview(idxs_a).cast('q', shape=(bp.length, 2))
        idxs = cvarray(shape=(bp.length, 2), itemsize=sizeof(Py_ssize_t), format="q")
    except MemoryError as exc:
        dtaidistancec_globals.dd_path_free(&bp)
        print("Cannot allocate memory for warping paths matrix.")
        raise exc
    for i in range(0, bp.length):
        idxs[i, 0] = bp.array[i].i
        idxs[i, 1] = bp.array[i].j
    dtaidistancec_globals.dd_path_free(&bp)
    return idxs

def loco_path_negativize(Py_ssize_t[:, :] path, seq_t[:, :] wps, int buffer, int inf_rows,
                              int inf_cols):
    dtaidistancec_loco.loco_path_negativize(&path[0,0], path.shape[0], &wps[0,0], wps.shape[0]-inf_rows, wps.shape[1]-inf_cols, buffer, inf_rows, inf_cols)

def loco_wps_argmax(seq_t[:, :] wps, Py_ssize_t[:] idxs, int n):
    l = wps.shape[0] * wps.shape[1]
    dtaidistancec_loco.loco_wps_argmax(&wps[0,0], l, &idxs[0], n)



# -*- coding: UTF-8 -*-
"""
dtaidistance.postprocessing
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Postprocessing similarity measures (e.g. paths).

:author: Wannes Meert
:copyright: Copyright 2025 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from .innerdistance import inner_dist_fns
from .preprocessing import derivative


def distance_from_path(ts1, ts2, path, inner_dist="squared euclidean",
                       relaxed=True):
    """Compute the distance defined by a given warping path.

    If relaxation is true, the difference between two points that are
    warped to each other is computed more relaxed to take into account
    the shape of the time series. If a point appears in a time series
    where the series is steep (i.e., a high derivative), then where the
    timepoint was sampled has a larger impact than for a flat part. To
    compensate for this, we assume that the time point could have been
    anywhere in the time range [t-Delta_t/2,t+Delta_t/2], where Delta_t
    is the timestep (since we do not take explicit timepoints, here
    Delta_t = 1). The smallest difference between all possible values
    in this time range is used.

    :param ts1: The first time series
    :param ts2: The first time series
    :param path: A list of points (index_ts1, index_ts2)
    :param inner_dist: One of :class:`InnerDistType`
    :param relaxed: Take the shape of the time series into account
    """
    inner_dist, result, _ = inner_dist_fns(inner_dist)
    if relaxed:
        ts1d = derivative(ts1)
        ts2d = derivative(ts2)
    else:
        ts1d, ts2d = None, None
    dist = 0
    for point in path:
        i_f, i_t = point
        if not relaxed:
            dist += inner_dist(ts1[i_f], ts2[i_t])
            continue

        deriv_1, deriv_2 = ts1d[i_f], ts2d[i_t]
        ts1_1, ts1_2 = ts1[i_f] - deriv_1 * 0.5, ts1[i_f] + deriv_1 * 0.5
        ts2_1, ts2_2 = ts2[i_t] - deriv_2 * 0.5, ts2[i_t] + deriv_2 * 0.5
        dist += min(inner_dist(ts1_1, ts2_1),
                    inner_dist(ts1_1, ts2_2),
                    inner_dist(ts1_2, ts2_1),
                    inner_dist(ts1_2, ts2_2))
    return result(dist)


def get_points_in_path_with_min_diff(ts1, ts2, path, min_diff, relaxed=True):
    """Get all points in the path where the difference between the two
    points that are warped to each other is larger than the given
    minimum difference.

    For the relaxed version, see :method:`distance_from_path`.
    """
    points = []
    for point in path:
        i_f, i_t = point
        diff = abs(ts1[i_f] - ts2[i_t])
        if diff > min_diff:
            if not relaxed:
                points.append(point)
                continue
            # Be more relaxed about points with a high derivative, sampling has a large impact there
            if 0 < i_f < len(ts1) - 1:
                deriv_ref = ((ts1[i_f] - ts1[i_f - 1]) + (ts1[i_f + 1] - ts1[i_f - 1]) / 2) / 2
            else:
                deriv_ref = 1
            if 0 < i_t < len(ts2) - 1:
                deriv_seg = ((ts2[i_t] - ts2[i_t - 1]) + (ts2[i_t + 1] - ts2[i_t - 1]) / 2) / 2
            else:
                deriv_seg = 1
            ts1_1, ts1_2 = ts1[i_f] - deriv_ref * 0.5, ts1[i_f] + deriv_ref * 0.5
            ts2_1, ts2_2 = ts2[i_t] - deriv_seg * 0.5, ts2[i_t] + deriv_seg * 0.5
            diff = min(abs(ts1_1-ts2_1),
                       abs(ts1_1-ts2_2),
                       abs(ts1_2-ts2_1),
                       abs(ts1_2-ts2_2))
            if diff > min_diff:
                points.append(point)
    return points



# -*- coding: UTF-8 -*-
"""
dtaidistance.util
~~~~~~~~~~~~~~~~~

Utility functions for DTAIDistance.

:author: Wannes Meert
:copyright: Copyright 2017-2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import os
import sys
import csv
import logging
from array import array
from pathlib import Path
import tempfile
import heapq
import time
from bisect import bisect_left, bisect_right
from enum import Enum


logger = logging.getLogger("be.kuleuven.dtai.distance")


try:
    import numpy as np
except ImportError:
    np = None

try:
    from . import dtw_cc
except ImportError:
    dtw_cc = None

try:
    from . import dtw_cc_omp
except ImportError:
    dtw_cc_omp = None

try:
    from . import dtw_cc_numpy
except ImportError:
    dtw_cc_numpy = None
except ValueError as exc:
    logger.warning('Warning: loading library to link with numpy returned an error')
    logger.warning(exc)
    dtw_cc_numpy = None


dtaidistance_dir = os.path.abspath(os.path.dirname(__file__))


def try_import_c(verbose=False):
    is_complete = True
    msgs = []
    global dtw_cc
    global dtw_cc_omp
    global dtw_cc_numpy
    try:
        from . import dtw_cc
    except Exception as exc:
        print('Cannot import C-based library (dtw_cc)')
        msgs.append('Cannot import C-based library (dtw_cc)')
        msgs.append(str(exc))
        dtw_cc = None
        is_complete = False
    try:
        from . import dtw_cc_omp
    except Exception as exc:
        print('Cannot import OMP-based library (dtw_cc_omp)')
        msgs.append('Cannot import OMP-based library (dtw_cc_omp)')
        msgs.append(str(exc))
        dtw_cc_omp = None
        is_complete = False
    try:
        from . import dtw_cc_numpy
    except Exception as exc:
        print('Cannot import Numpy-based library (dtw_cc_numpy)')
        msgs.append('Cannot import Numpy-based library (dtw_cc_numpy)')
        msgs.append(str(exc))
        dtw_cc_numpy = None
        is_complete = False
    try:
        import numpy
        msgs.append('Numpy version: {}'.format(numpy.__version__))
    except Exception as exc:
        print('Cannot import Numpy (optional dependency)')
        msgs.append('Cannot import Numpy (optional dependency)')
        msgs.append(str(exc))
    try:
        import matplotlib
        msgs.append('Matplotlib version: {}'.format(matplotlib.__version__))
    except Exception as exc:
        print('Cannot import Matplotlib (optional dependency)')
        msgs.append('Cannot import Matplotlib (optional dependency)')
        msgs.append(str(exc))
    try:
        import scipy
        msgs.append('Scipy version: {}'.format(scipy.__version__))
    except Exception as exc:
        print('Cannot import SciPy (optional dependency)')
        msgs.append('Cannot import SciPy (optional dependency)')
        msgs.append(str(exc))
    if not is_complete:
        print('\nNot all libraries are available in your installation. ')
        print('You can rerun the compilation from source or pip install in verbose mode:\n'
              'pip install -vvv --upgrade --force-reinstall --no-deps --no-binary dtaidistance dtaidistance')
        print('In case you need to use an older version of numpy, compile against your current installation:\n'
              'pip install -vvv --upgrade --force-reinstall --no-deps --no-build-isolation '
              '--no-binary dtaidistance dtaidistance')
        print('\nShare the following information when submitting a bug report:')
    elif verbose:
        print('All ok ...')
    if not is_complete or verbose:
        print('== Packages ==')
        for msg in msgs:
            print('- {}'.format(msg))
        print('== System information ==')
        import sys
        print(sys.implementation)
        print('== Compilation information ==')
        try:
            import pkgutil
            logtxt = pkgutil.get_data(__name__, "compilation.log")
            print(logtxt.decode())
        except Exception as exc:
            print('Could not read compilation.log')
            print(exc)
        print('')
        print('==')
    return is_complete


def prepare_directory(directory=None):
    """Prepare the given directory, create it if necessary.
    If no directory is given, a new directory will be created in the system's temp directory.
    """
    if directory is not None:
        directory = Path(directory)
        if not directory.exists():
            directory.mkdir(parents=True)
        logger.debug("Using directory: {}".format(directory))
        return Path(directory)
    directory = tempfile.mkdtemp(prefix="dtaidistance_")
    logger.debug("Using directory: {}".format(directory))
    return Path(directory)


def read_substitution_matrix(file):
    """Read substitution matrix from file.

    Comments starting with # and newlines are allowed anywhere
    in the file.
    
    :return: A dictionary mapping tuples of symbols to their weight.
    """

    def strip_comments(reader):
        for line in reader:
            if not line.rstrip() or line[0] == '#':
                continue
            yield line.rstrip()

    matrix = dict()
    with open(file) as f:
        reader = csv.reader(strip_comments(f), delimiter=" ", skipinitialspace=True)
        line = next(reader)
        idx = {i: symbol for i, symbol in enumerate(line)}
        for line in reader:
            symbol = line[0]
            for j, value in enumerate(line[1:]):
                matrix[(idx[j], symbol)] = float(value)
    return matrix


def detect_ndim(s):
    if np is not None and isinstance(s, np.ndarray):
        return s.ndim
    if type(s) is list and len(s) > 0:
        return detect_ndim(s[0]) + 1
    if type(s) in [int, float]:
        return 0
    return None


class SeriesContainer:
    def __init__(self, series, support_ndim=True):
        """Container for a list of series.

        This wrapper class knows how to deal with multiple types of datastructures to represent
        a sequence of sequences.

        For 1 dimensional time series (e.g. [1,2,3,4]):
        - List[array.array]
        - List[numpy.array] (array is 1 dimensional)
        - List[List]
        - numpy.array (2 dimensional)
        - numpy.matrix

        For n-dimensional time series (e.g. [[1,2],[3,4],[5,6]]):
        - List[numpy.array] (array is 2 dimensional)
        - List[List[List]]
        - numpy.array (3 dimensional)

        When using the C-based extensions, the data is automatically verified and converted.
        """
        self.support_ndim = support_ndim
        # Always detect the dimensionality of the time series, even if support_ndim is false
        self.detected_ndim = False
        if isinstance(series, SeriesContainer):
            self.series = series.series
        elif np is not None and isinstance(series, np.ndarray):
            # A np.matrix always returns a 2D array, also if you select one row (to be consistent
            # and always be a matrix datastructure). The methods in this toolbox expect a
            # 1D array thus we need to convert to a 1D or 2D array. This is taken care by asarray
            self.series = np.asarray(series, order="C")
            if self.series.ndim > 2:
                if not self.support_ndim:
                    raise Exception('N-dimensional series are not supported '
                                    '(series.ndim = {}) > 2'.format(self.series.ndim))
                self.detected_ndim = len(self.series[0, 0])
            else:
                self.detected_ndim = 1
        elif type(series) in [set, tuple, list]:
            self.series = list(series)
            if np is not None and isinstance(self.series[0], np.ndarray):
                if self.series[0].ndim > 1:
                    if not self.support_ndim:
                        raise Exception('N-dimensional series are not supported '
                                        '(series[0].ndim = {}) > 1'.format(self.series[0].ndim))
                    self.detected_ndim = len(self.series[0][0])
                else:
                    self.detected_ndim = 1
            elif type(series[0]) in [list, tuple]:
                if type(series[0][0]) in [list, tuple]:
                    self.detected_ndim = len(self.series[0][0])
                else:
                    self.detected_ndim = 1
        else:
            self.series = series

    def set_detected_ndim(self, ndim):
        if ndim is None:
            return
        self.detected_ndim = ndim

    def c_data_compat(self):
        """Return a datastructure that the C-component knows how to handle.
        The method tries to avoid copying or reallocating memory.

        :return: Either a list of buffers or a two-dimensional buffer. The
            buffers are guaranteed to be C-contiguous and can thus be used
            as regular pointer-based arrays in C.
        """
        if dtw_cc is None:
            raise Exception('C library not loaded')
        if type(self.series) == list:
            for i in range(len(self.series)):
                serie = self.series[i]
                if np is not None and isinstance(serie, np.ndarray):
                    if not self.support_ndim and serie.ndim != 1:
                        raise Exception('N-dimensional arrays are not supported (serie.ndim = {})'.format(serie.ndim))
                    if not serie.flags.c_contiguous:
                        serie = np.asarray(serie, order="C")
                        self.series[i] = serie
                elif isinstance(serie, array):
                    pass
                else:
                    raise Exception(
                        "Type of series not supported, "
                        "expected numpy.array or array.array but got {}".format(
                            type(serie)
                        )
                    )
            return dtw_cc.dtw_series_from_data(self.series)
        elif np is not None and isinstance(self.series, np.ndarray):
            if not self.series.flags.c_contiguous:
                logger.warning("Numpy array not C contiguous, copying data.")
                self.series = self.series.copy(order="C")
            if not self.support_ndim and self.series.ndim > 2:
                raise Exception(f'N-dimensional series are not supported (series.ndim = {self.series.ndim})')
            if dtw_cc_numpy is None:
                logger.warning("DTAIDistance C-extension for Numpy is not available. Proceeding anyway.")
                return dtw_cc.dtw_series_from_data(self.series)
            elif len(self.series.shape) == 3:
                return dtw_cc_numpy.dtw_series_from_numpy_ndim(self.series)
            else:
                return dtw_cc_numpy.dtw_series_from_numpy(self.series)
        return dtw_cc.dtw_series_from_data(self.series)

    def get_max_y(self):
        max_y = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_y = max(np.max(self.series), abs(np.min(self.series)))
        else:
            for serie in self.series:
                max_y = max(max_y, np.max(serie), abs(np.min(serie)))
        return max_y

    def get_max_min_y(self):
        min_y, max_y = 0, 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            min_y = np.min(self.series)
            max_y = np.max(self.series)
        else:
            for serie in self.series:
                min_y = min(min_y, np.min(serie))
                max_y = max(max_y, np.max(serie))
        return min_y, max_y

    def get_max_length(self):
        max_length = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_length = self.series.shape[1]
        else:
            for serie in self.series:
                max_length = max(max_length, len(serie))
        return max_length

    def get_avg_length(self):
        max_length = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_length = self.series.shape[1]
        else:
            for serie in self.series:
                max_length += len(serie)
            max_length /= len(self.series)
        return max_length

    def __getitem__(self, item):
        return self.series[item]

    def __len__(self):
        return len(self.series)

    def __str__(self):
        return "SeriesContainer:\n{}".format(self.series)

    @staticmethod
    def wrap(series, support_ndim=True):
        if isinstance(series, SeriesContainer):
            series.support_ndim = support_ndim
            return series
        return SeriesContainer(series, support_ndim=support_ndim)


def recompile():
    import subprocess as sp

    sp.run([sys.executable, "setup.py", "build_ext", "--inplace"], cwd=dtaidistance_dir)


def argmin(a):
    imin, vmin = 0, float("inf")
    for i, v in enumerate(a):
        if v < vmin:
            imin, vmin = i, v
    return imin


def argmax(a):
    imax, vmax = 0, float("-inf")
    for i, v in enumerate(a):
        if v > vmax:
            imax, vmax = i, v
    return imax


class DetectKnee:
    def __init__(self, alpha=0.3, invert=False, alpha_onlyvar=0.0,
                 thr_value=None, verbose=False):
        """Exponential Weighted Moving Average (EWMA) based knee detection.

        Useful to detect when values start increasing at an increased rate.

        Based on:
        https://cseweb.ucsd.edu//~snoeren/papers/plush-usenix06.pdf

        :param alpha: EWMA parameter, in [0,1]
            Low values prefer old values, high values prefer recent values.
        :param invert: Find a knee in a curve that decreases.
            If false, an increasing curve is expected.
        :param alpha_onlyvar: EWMA parameter, in [0,1]
            The `dostop` method has an option to not update but only update
            the variance. The variance is reduced with the value.
            Low values prefers to keep the variance estimate, high value reduce the variance faster.
        """
        self.cnt = 0  # Number of data points seen
        self.cnt_onlyvar = 0
        self.min_points = 3  # Minimal number of data points to see before stopping
        self.thr_value = thr_value if thr_value is not None else 0
        self.arrvar_fraction = 4
        self.alpha = alpha  # EWMA parameter
        self.alpha_onlyvar = alpha_onlyvar
        self.arr = None
        self.arrvar = None
        self.max_thr = None
        self.invert = invert
        self.invert_value = None

        self.verbose = verbose
        if self.verbose:
            print(f"Search knee with alpha={alpha}")

    def dostop(self, value, only_var=False):
        """Check whether a knee has occurred for a new value.

        :param value: New value in the curve
        :param only_var: Only shrink the expected variance, do not update the average value.
        :return: Is a knee detected
        """
        if self.arr is None:
            if self.invert:
                self.invert_value = value
                self.thr_value = value - self.thr_value
                value = 0
            self.arr = value
            self.arrvar = 0
            if self.verbose:
                print(f"Detecting knee: threshold value={self.thr_value:.4f}")
            return False
        elif self.invert:
            value = self.invert_value - value
        rvalue = False
        self.max_thr = self.arr + self.arrvar_fraction * self.arrvar
        # We need to see at least min_points instances to compute a reasonable arrvar
        if self.cnt >= self.min_points and value > self.thr_value and value > self.max_thr:
            rvalue = True

        if only_var:
            self.arrvar = (1.0 - self.alpha_onlyvar) * self.arrvar
            self.cnt_onlyvar += 1
        else:
            self.arrvar = self.alpha * max(0, value - self.arr) + (1.0 - self.alpha) * self.arrvar
            self.arr = self.alpha * value + (1.0 - self.alpha) * self.arr

        if self.verbose and (not only_var or rvalue is True):
            if self.invert:
                pvalue = f"{value:.4f} ({self.invert_value:.4f}-{self.invert_value - value:.4f})"
            else:
                pvalue = f"{value:.4f}"
            if self.cnt < self.min_points:
                print(f"Detected knee: Not enough points, {self.cnt} < {self.min_points}, "
                      f"{pvalue} > {self.max_thr:.4f} = {self.arr:.4f} + {self.arrvar_fraction} * {self.arrvar:.4f}")
            else:
                print(f"Detected knee: {self.cnt} ({self.cnt_onlyvar}) points, {pvalue} > {self.max_thr:.4f} = "
                      f"{self.arr:.4f} + {self.arrvar_fraction} * {self.arrvar:.4f} -> {rvalue}")

        if not only_var:
            self.cnt += 1
        return rvalue


class SortedList:
    def __init__(self, values):
        self._l = sorted(values)

    def __len__(self):
        return self._l.__len__()

    def __iter__(self):
        return self._l.__iter__()

    def __getitem__(self, item):
        return self._l.__getitem__(item)

    def remove(self, x):
        """Remove the value equal to x."""
        try:
            i = self.index(x)
            del self._l[i]
        except ValueError:
            pass

    def index(self, x):
        """Locate the leftmost value exactly equal to x."""
        i = bisect_left(self._l, x)
        if i != len(self._l) and self._l[i] == x:
            return i
        raise ValueError

    def find_lt(self, x):
        """Find rightmost value less than x."""
        i = bisect_left(self._l, x)
        if i:
            return self._l[i - 1]
        raise ValueError

    def find_le(self, x):
        """Find rightmost value less than or equal to x."""
        i = bisect_right(self._l, x)
        if i:
            return self._l[i - 1]
        raise ValueError

    def find_gt(self, x):
        """Find leftmost value greater than x."""
        i = bisect_right(self._l, x)
        if i != len(self._l):
            return self._l[i]
        raise ValueError

    def find_ge(self, x):
        """Find leftmost item greater than or equal to x."""
        i = bisect_left(self._l, x)
        if i != len(self._l):
            return self._l[i]
        raise ValueError

    def __str__(self):
        return str(self._l)

    def __repr__(self):
        return str(self._l)


class DDType(str, Enum):

    def to_int(self):
        return list(self.__class__).index(self)

    @classmethod
    def from_int(cls, value):
        return list(cls)[value]

    @classmethod
    def wrap(cls, val, default_val=None):
        if val is None:
            val = default_val
        if isinstance(val, cls):
            return val
        elif type(val) is str:
            try:
                return cls(val)
            except ValueError as exc:
                pass
        elif type(val) is int:
            try:
                return cls.from_int(val)
            except IndexError as exc:
                pass
        raise ValueError(f'Value not supported for {cls.__name__}: {val}')



# -*- coding: UTF-8 -*-
"""
dtaidistance.subsequence.dtw
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(requires version 2.3.0 or higher)

DTW-based subsequence matching.

:author: Wannes Meert
:copyright: Copyright 2021-2023 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""

from .subsequencealignment import *
from .localconcurrences import *
from .subsequencesearch import *





cdef extern from "dd_globals.h":
    ctypedef double seq_t

    ctypedef enum StepType:
        TypeI,
        TypeIII

    ctypedef struct DDPathEntry:
        Py_ssize_t i
        Py_ssize_t j

    ctypedef struct DDPath:
        DDPathEntry * array
        Py_ssize_t length
        Py_ssize_t capacity
        seq_t distance

    void dd_path_free(DDPath *a)




import math
import numpy as np

from ..util import SeriesContainer
from ..subsequence.subsequencealignment import subsequence_alignment
from ..exceptions import MatplotlibException
from ..similarity import distance_to_similarity


class SymbolAlignment:
    def __init__(self, codebook, maxcompression=0.5, maxexpansion=2):
        """Translate a time series with continuous values to a list of discrete
        symbols based on motifs in a codebook.

        :param codebook: List of motifs.
        :param maxcompression: Maximally allowed compression of a codeword for it to be recognized
        :param maxexpansion: Maximally allowed expansion of a codeword for it to be recognized
        """
        self.codebook = codebook
        self.maxcompression = maxcompression
        self.maxexpansion = maxexpansion
        self.use_c = False
        self.symbols = None
        self._agg_fn = self.agg_min
        self._agg_args = None

    def set_agg_min(self):
        """Set mininum aggregation for the align2 method."""
        self._agg_fn = self.agg_min
        self._agg_args = None

    def agg_min(self, patterns, max_value):
        return np.argmin(patterns, axis=2).astype(int)

    def set_agg_prob(self, window=10):
        """Set probabilistic aggregation for the align2 method."""
        self._agg_fn = self.agg_prob
        self._agg_args = (window,)

    def agg_prob(self, patterns, max_value):
        window = self._agg_args[0]
        tlen = patterns.shape[1]
        patterns = distance_to_similarity(patterns, r=max_value, method='exponential')
        # Softmax
        exp_x = np.exp(patterns - np.max(patterns, axis=2)[:, :, np.newaxis])
        logprobs = np.log(exp_x / np.sum(exp_x, axis=2)[:, :, np.newaxis])
        best = np.zeros((patterns.shape[0], tlen), dtype=int)
        for ti in range(window):
            best[:, ti] = np.argmax(logprobs[:, :ti, :].sum(axis=1), axis=1)
        for ti in range(window, tlen-window):
            best[:, ti] = np.argmax(logprobs[:, ti:ti+window, :].sum(axis=1), axis=1)
        for ti in range(tlen-window, tlen):
            best[:, ti] = np.argmax(logprobs[:, ti:, :].sum(axis=1), axis=1)
        return best

    def align2(self, series):
        """Perform alignment.

        For each time point, select the best matching motif and aggregate the results.

        :param series: List of time series or a numpy array
        """
        sc = SeriesContainer(series)

        patterns = np.zeros((len(sc), sc.get_max_length(), len(self.codebook) + 1))
        patterns[:, :, :] = np.inf
        max_value = 0
        for sidx in range(len(sc)):
            for midx in range(len(self.codebook)):
                medoidd = np.array(self.codebook[midx])
                sa = subsequence_alignment(medoidd, sc[sidx], use_c=self.use_c)
                for match in sa.kbest_matches(k=None,
                                              minlength=math.floor(len(medoidd)*self.maxcompression),
                                              maxlength=math.ceil(len(medoidd)*self.maxexpansion)):
                    patterns[sidx, match.segment[0]:match.segment[1]+1, midx] = match.value
                    max_value = max(max_value, match.value)
        patterns[:, :, len(self.codebook)] = 0
        patterns[:, :, len(self.codebook)] = np.max(patterns) + 1
        best_patterns = self._agg_fn(patterns, max_value)
        self.symbols = best_patterns
        return best_patterns

    def align(self, series, max_rangefactor=None, detectknee_alpha=None, max_overlap=None):
        """Perform alignment.

        Only one motif is matched to a subsequence. No aggregation within the same subsequence.

        :param series: List of time series or a numpy array
        :param max_rangefactor: the range between the first (best) match and the last match
            can be at most a factor of ``maxrangefactor``. For example, if the first match has
            value v_f, then the last match has a value ``v_l < v_f*maxfactorrange``.
        :param max_overlap: Maximal overlap when matching codewords.
            If not given, this is based on maxcompression and maxexpansion.
        """
        # Inspired on the Matching Pursuit algorithm.
        sc = SeriesContainer(series)
        noword = len(self.codebook)
        best_patterns = np.full(series.shape, noword, dtype=int)
        if max_overlap is None:
            max_overlap = max(self.maxcompression, 1./self.maxexpansion)

        for sidx in range(len(sc)):
            curseries = sc[sidx].copy()
            patterns = []
            max_value = 0
            for midx in range(len(self.codebook)):
                medoidd = np.array(self.codebook[midx])
                sa = subsequence_alignment(medoidd, curseries, use_c=self.use_c)
                if max_rangefactor is not None:
                    itr = sa.best_matches(max_rangefactor=max_rangefactor,
                                          minlength=math.floor(len(medoidd) * self.maxcompression),
                                          maxlength=math.ceil(len(medoidd) * self.maxexpansion))
                else:
                    itr = sa.best_matches_knee(alpha=detectknee_alpha,
                                          minlength=math.floor(len(medoidd) * self.maxcompression),
                                          maxlength=math.ceil(len(medoidd) * self.maxexpansion))
                for match in itr:
                    patterns.append((midx, match.segment[0], match.segment[1]+1,
                                     curseries[match.segment[0]:match.segment[1]+1], match.value))
                    max_value = max(max_value, match.value)
            # print(f"Series {sidx}: found {len(patterns)} patterns")
            # D = np.zeros((len(patterns), len(curseries)))
            # for i, (_, bi, ei, ts) in enumerate(patterns):
            #     D[i, bi:ei] = ts # should be normalized for matching pursuit if we use the factors
            D = np.zeros(len(patterns))
            L = np.zeros(len(patterns), dtype=int)
            B = np.zeros(len(patterns), dtype=int)
            E = np.zeros(len(patterns), dtype=int)
            for i, (_, bi, ei, ts, v) in enumerate(patterns):
                D[i] = v
                L[i] = ei - bi + 1
                B[i] = bi
                E[i] = ei + 1  # Use next position for length and overlap computation
            # Trade-off between length and similarity
            S = distance_to_similarity(D, r=max_value, method='exponential') * L

            bestpatvalue = np.inf
            its = 0
            while bestpatvalue > 0:
                # print(f"Iteration {sidx} -- {its}")
                # dotprod = np.einsum('j,kj->k', curseries, D)
                # bestpatidx = np.argmax(dotprod)
                # bestpatvalue = dotprod[bestpatidx]
                bestpatidx = np.argmax(S)
                bestpatvalue = S[bestpatidx]
                if bestpatvalue == 0:
                    break
                pattern = patterns[bestpatidx]
                freeidxs = best_patterns[sidx, pattern[1]:pattern[2]] == noword
                best_patterns[sidx, pattern[1]:pattern[2]][freeidxs] = pattern[0]
                # Remove patterns with more overlap than max_overlap
                overlaps = (np.maximum(0, np.minimum(E[bestpatidx], E) -
                                          np.maximum(B[bestpatidx], B)) / L[bestpatidx]) > max_overlap
                S[overlaps] = 0
                # curseries[pattern[1]:pattern[2]] = 0
                # D[bestpatidx, :] = 0
                S[bestpatidx] = 0
                its += 1

        self.symbols = best_patterns
        return best_patterns

    def align_fast(self, *args, **kwargs):
        """See :meth:`align`."""
        use_c = self.use_c
        self.use_c = True
        result = self.align(*args, **kwargs)
        self.use_c = use_c
        return result

    def hangover(self, symbols, threshold=4):
        """Hangover filter for symbols."""
        sequences = []
        sequences_idx = []
        for r in range(symbols.shape[0]):
            sequence = []
            sequence_idx = []
            lastval = None
            lastcnt = 0
            firstidx = None
            lastsaved = None
            for c, v in enumerate(symbols[r, :]):
                if v != lastval:
                    if lastcnt > threshold and lastval != lastsaved:
                        sequence.append(lastval + 1)  # cannot be zero
                        sequence_idx.append((firstidx, c))
                        lastsaved = lastval
                    lastval = v
                    lastcnt = 0
                    firstidx = c
                else:
                    lastcnt += 1
            sequences.append(sequence)
            sequences_idx.append(sequence_idx)
        return sequences, sequences_idx

    def plot(self, series, sequences, sequences_idx, ylabels=None, filename=None, figsize=None,
             xlimits=None, symbollabels=None):
        try:
            import matplotlib.pyplot as plt
            from matplotlib import gridspec
        except ImportError:
            raise MatplotlibException("No matplotlib available")
        if figsize is None:
            figsize = (12, 8)
        sc = SeriesContainer(series)
        fig, axs = plt.subplots(nrows=len(sc), ncols=1, sharex='all', sharey='col', figsize=figsize)
        if len(sc) == 1:
            axs = [axs]
        for r in range(series.shape[0]):
            # avg_value = series[r, :].mean()
            avg_value = series.min() + (series.max() - series.min()) / 2
            if xlimits is None:
                axs[r].plot(series[r, :])
            else:
                axs[r].plot(series[r, xlimits[0]:xlimits[1]])
            if ylabels is not None:
                axs[r].set_ylabel(f"L={ylabels[r]}")
            for symbol, (fidx, lidx) in zip(sequences[r], sequences_idx[r]):
                if xlimits is None:
                    delta = 0
                elif xlimits[0] <= fidx <= xlimits[1]:
                    delta = xlimits[0]
                else:
                    continue
                axs[r].vlines(fidx-delta, series.min(), series.max(), colors='k', alpha=0.2)
                if symbollabels is not None:
                    symbol = symbollabels(symbol)
                axs[r].text(fidx-delta, avg_value, str(symbol), alpha=0.5)
        if filename is not None:
            fig.savefig(filename)
            plt.close(fig)
        else:
            return fig, axs
